#ifndef CLEANERREADER_CPP_UNIVERSITY_OF_OREGON_NIC
#define CLEANERREADER_CPP_UNIVERSITY_OF_OREGON_NIC

#include "CleanerReader.h"

// DEFINE FORM DATA NAMES HERE -------------------------------------------------
#define N_CLEANER_PHRASES 7
char * CLEANER_PHRASES[N_CLEANER_PHRASES] = {
  "Process",
  "Source data location",
  "File type",
  "Orientation",
  "Channels",
  "Observations",
  "N processors"
};



CleanerReader::CleanerReader( istream& inStream )
:   Reader( inStream )
{
    phrase = CLEANER_PHRASES;
    n_phrases = N_CLEANER_PHRASES;
    n_procs = 1;  // If MPI, this is preset, if OMP, it needs to be set

    data_src = NULL;
}

CleanerReader::~CleanerReader()
{
    if ( data_src != NULL) delete [] data_src;
}

void CleanerReader::validate_read()
{
    cout << "VALIDATE READING OF CLEANER FILE" << endl;
    cout << "data source file: " << data_src << endl;
    cout << "n observations  : "<< obs << endl;
    cout << "n channels      : "<< ch << endl;
    cout << "orientation     : "<< orient << endl;
}

void CleanerReader::get_value()
{
    char * s;
    char c;
    int n;

    switch (current_phrase) 
    {
    case 0 :
        get_process_type(current_phrase);
        break;

    case 1 :
        s = get_str(value);
        n = strlen(s);

        // do we need to delete this first? cch
        data_src = new char[n+25];
        sprintf(data_src, "%s", s);
        break;

    case 2 :
        type = get_char(value);
        break;

    case 3 :
        orient = get_char(value);
        break;

    case 4 :
        ch = get_int(value);
        break;

    case 5 :
        obs = get_int(value);
        break;

#ifndef __DISTRIBUTED
    case 6 :
        n_procs = get_int(value);
        break;
#endif

    default :
        break;
    }

}

/*******************************************************************************
    The distributed data load is a little complex. If the data is loaded from a 
    binary file in column-major format then we can simply distribute the load, 
    if the file is not in this format, then a single processor should read and
    distribute the array.

    Each processor reads n observations from the data array, where 
    n = n_obs/n_cpus. If the remainder of n_obs/n_cpus is m, where m is 
    non-zero, each of the first m processors gets an addition observation. To 
    ensure that the observations are disjoint, we compute an offset that 
    identifies the location of each processors first observation. The 
    computation is as follows:

    if processor i < m, then obs_ = n+1 for that processor and all processor
                        j < i the offset for i is i*obs_
    if processor i >= m, then obs_ = n so, for that processor, the offset is
                        the sum of obs_ for the first m processors (n+1)*m
                        and the sum of the next i-m processors n*(i-m), or
                        offset = n*m + m + n*i - n*m = m + n*i
   
   ****************************************************************************/
nic_matrix* CleanerReader::load_matrix( int rank, int n_cpus )
{
    nic_matrix* data = 0;
    int total_rows = ch;
    int total_columns = obs; // save the original value and reset it at the end
    int rem  = obs % n_cpus;
    int offset;

    // Object for loading the data in the processor
    PartitionedDataset              partitionedData;
    int columns = obs / n_cpus;

    // rem is the number of processors that must take an additional observation
    if (rank < rem) ++columns;

    // compute the offset (see above for logic)
    offset = rank * columns;
    if (rank >= rem) offset += rem;
    offset *= ch;

    // I am creating this for someone... who?
    data = new nic_matrix(ch, columns, total_rows, total_columns, 
                                  rank, n_cpus);

    type = std::toupper( type );
    orient = std::toupper( orient );
                            
    if ( ( type == 'B' ) || ( type == 'L' ) || ( type == 'P' ) )
    {  // binary file
        if ( orient == 'C' )
        {
            partitionedData.loadFile( 
                rank, data_src, data, 
                ch, columns,rem, obs, 
                n_cpus, type, orient );
        }
        else if ( orient == 'O' || orient == 'o' )
        {
            cerr << "orientation=O is not yet verified, and therefore, not yet supported" << endl;
            assert( ( orient == 'O' || orient == 'o' ) );

#ifdef VERIFYME
            if (rank == 0)
            {
                DBinaryMatrixT<double> mat_reader(
                    data, data_src, offset, total_columns);
            } 
            else 
            {
                MPI_Status status;
                // receive and unpack data, store it in the matrix
                MPI_Recv(
                    data->data, data->size, 
                    MPI_DOUBLE, 0, 1, MPI_COMM_WORLD, &status);
            }
#endif
        }   // checking binary files
    }
    else if ( type == 'A' || type == 'a' )
    {
        cerr << "type=A is not yet verified, and therefore, not yet supported" << endl;
        assert( ( type == 'A' || type == 'a' ) );
        
#ifdef VERIFYME
        if (orient == 'C' || orient == 'c') 
        {
            if (rank == 0) 
            {
                DASCIIMatrix<double> mat_reader(
                    data, data_src, offset, total_columns);
            } 
            else 
            {
                // wait on load
                MPI_Recv(
                    data->data, data->size, MPI_DOUBLE, 
                    0, 1, MPI_COMM_WORLD, &status);
            }
        }
        else if (orient == 'O' || orient == 'o')
        {
            cerr << "orient == 'O' is not yet verified, and therefore, not yet supported" << endl;
            assert( orient == 'O' || orient == 'o' );
            if (rank == 0) 
            {
                DASCIIMatrixT<double> mat_reader(
                    data, data_src, offset, total_columns);
            } 
            else 
            {
                // wait on load
                MPI_Recv(
                    data->data, data->size, 
                    MPI_DOUBLE, 0, 1, MPI_COMM_WORLD, &status);
            }
        }
#endif
    } 
    else 
    {
      // error 
    }

    // reset obs to the full value

    // whomever calls me has to deallocate data
    return data;
}

void CleanerReader::get_process_type( int current_phrase )
{
    char * s = get_str(value);
  
    if (!strcmp(s, "FastICA")) process = FASTICA;
    else if (!strcmp(s, "Infomax")) process = INFOMAX;
    else if (!strcmp(s, "PCA")) process = PCA;
    else if (!strcmp(s, "None")) process = NONE;
    else 
    {
      std::cout << "ERROR: Process type defined in the signal cleaning form ";
      std::cout << " was invalid." << std::endl;
      exit(-100);
    }
}

#endif
// CLEANERREADER_CPP_UNIVERSITY_OF_OREGON_NIC
