#include "DataReader.h"
#include "DataWriter.h"
#include "SimpleCLParser.h"
#include <string>
#include <fstream>
#include "version.h"

using namespace std;

void setupParser( SimpleCLParser& parser )
{
    bool isRequired = true;
    bool isNotRequired = false;

    // usage and copyright info
    string usage = "\nHiPerSAT Convert v. 1.0 (r";
    usage += hipersatVersionString();
    usage +=  ")\n\n";
    usage += NIC_COPYRIGHT;
    parser.addUsageHeader( usage );


    parser.addOption( 
        SimpleCLParser::STRING,
        "-i",
        "input file", 
        isRequired );

    parser.addOption(
        SimpleCLParser::STRING,
        "-o",
        "output file",
        isRequired );

    parser.addOption(
        SimpleCLParser::STRING,
        "-if",
        "input format (big, little, native, text)",
        isRequired );

    parser.addOption(
        SimpleCLParser::STRING,
        "-of",
        "output format (big, little, native, text)",
        isRequired );

    parser.addOption(
        SimpleCLParser::FLAG,
        "-h",
        "show usage",
        isNotRequired );

    parser.addOption(
        SimpleCLParser::INTEGER,
        "-c",
        "number of channels",
        isNotRequired );


    parser.parse();
    if ( parser.readFlagOption( "-h" ) )
    {
        std::cout << parser.usage() << std::endl;
    }

    parser.checkRequired();

}

void exitOnError( const char* message )
{
    std::cerr << message << std::endl;
    std::cerr << "Exiting." << std::endl;
    exit( 1 );
}

ifstream* openInputStream( SimpleCLParser& parser )
{
    string fileName = parser.readStringOption( "-i" );
    bool isText = (parser.readStringOption( "-if" ) == "text");
    
    ifstream* newFile = new ifstream( 
        fileName.c_str(),
        isText ? ios::in : ios::binary );

    if ( !newFile )
    {
        string message( "Unable to open input file " );
        exitOnError( ( message + fileName ).c_str() );
    }
    return newFile;
}

ofstream* openOutputStream( SimpleCLParser& parser )
{
    string fileName = parser.readStringOption( "-o" );
    bool isText = (parser.readStringOption( "-of" ) == "text");
    
    ofstream* newFile = new ofstream( 
        fileName.c_str(),
        isText ? ios::out : ios::binary );

    if ( !newFile )
    {
        string message( "Unable to open output file " );
        exitOnError( ( message + fileName ).c_str() );
    }
    return newFile;
}


DataReader<double>* createReader( SimpleCLParser& parser, ifstream* in )
{
    string informat = parser.readStringOption( "-if" );
    DataReader<double>::Format format;
    DataReader<double>::Endian endian;
    if ( "big" == informat )
    {
        format = DataReader<double>::Binary;
        endian = DataReader<double>::BigEndian;
    }
    else if ( "little" == informat )
    {
        format = DataReader<double>::Binary;
        endian = DataReader<double>::LittleEndian;
    }
    else if ( "native" == informat )
    {
        format = DataReader<double>::Binary;
        endian = DataReader<double>::NativeEndian;
    }
    else if ( "text" == informat )
    {
        format = DataReader<double>::Ascii;
        endian = DataReader<double>::NativeEndian;
    }
    else
    {
        string message( "Unknown input file format " );
        exitOnError( ( message + informat ).c_str() );
    }

    return new DataReader<double>( *in, format, endian );
}

DataWriter<double>* createWriter( SimpleCLParser& parser, ofstream* out )
{
    string outformat = parser.readStringOption( "-of" );
    DataWriter<double>::Format format;
    DataWriter<double>::Endian endian;
    if ( "big" == outformat )
    {
        format = DataWriter<double>::Binary;
        endian = DataWriter<double>::BigEndian;
    }
    else if ( "little" == outformat )
    {
        format = DataWriter<double>::Binary;
        endian = DataWriter<double>::LittleEndian;
    }
    else if ( "native" == outformat )
    {
        format = DataWriter<double>::Binary;
        endian = DataWriter<double>::NativeEndian;
    }
    else if ( "text" == outformat )
    {
        format = DataWriter<double>::Ascii;
        endian = DataWriter<double>::NativeEndian;
        out->precision( 16 );
    }
    else
    {
        string message( "Unknown output file format " );
        exitOnError( ( message + outformat ).c_str() );
    }

    return new DataWriter<double>( *out, format, endian );
}


int main( int argc, char** argv )
{
    SimpleCLParser clParser( argc, argv );
    setupParser( clParser );
    cout << clParser.usageHeader() << endl;
    ifstream* inputFile = openInputStream( clParser );
    ofstream* outputFile = openOutputStream( clParser );
    DataReader<double>* reader = createReader( clParser, inputFile );
    DataWriter<double>* writer = createWriter( clParser, outputFile );

    double value;
    int count = 0;
    int channels;
    if ( clParser.readFlagOption( "-c" ) )
    {
        channels = clParser.readIntOption( "-c" );
    }
    else
    {
        channels = 10; // a default value for number of text columns
    }
    while( reader->read( value ) )
    {
        writer->write( value );
        ++count;
        if ( count == channels )
        {
            writer->linebreak();
            count = 0;
        }
    }

    delete reader;
    delete writer;
    delete inputFile;
    delete outputFile;
}
