#ifndef FASTICA_CPP_UNIVERSITY_OF_OREGON_NIC
#define FASTICA_CPP_UNIVERSITY_OF_OREGON_NIC

#include "FastICA.h"
#include <math.h>
#include "LAPACKWrapper.h"

FastICA::FastICA(
    nic_matrix* d, FastICAReader * reader, 
    nic_matrix* sph, unsigned long seed ) 
:   data(d),
    n_channels(d->rows), n_observations(d->columns),
    tolerance(reader->tolerance), max_iterations(reader->iterations),
    m_randNumGen( seed ),
    m_sph( sph ),
    #ifdef __DISTRIBUTED
    m_rb( MPI_COMM_WORLD )
    #else
    m_rb( 1 )
    #endif
{
    if ( 'r' == reader->initialization_type || 'R' == reader->initialization_type )
    {
        m_useRandomWeights = true;
        max_retries = reader->m_numRetries;
    }
    else
    {
        m_useRandomWeights = false;
        max_retries = 1;
    }
    nic_matrix::incrementReferenceCount( d );

    w = new nic_vector(n_channels);
    w_plus = new nic_vector(n_channels);
    w_temp = new nic_vector(n_channels);
    weights = new nic_matrix(n_channels, n_channels);

    init_matrix = new nic_matrix(n_channels, n_channels);
    BBt = new nic_matrix(n_channels, n_channels);

    error = 100;

    rank = d->rank;

    #ifdef __DISTRIBUTED
    total_size = d->total_size;
    #else
    total_size = d->columns;
    #endif 

    load_init_matrix(reader,sph);
    load_contrast(reader->contrast);

#ifdef __DISTRIBUTED
    if ( rank == 0 )
#endif
        std::cout << "Seed: " << seed << std::endl;
}

FastICA::~FastICA()
{
    nic_matrix::freeReference( data );

    delete w;
    delete w_plus;
    delete w_temp;
    delete weights;
    delete init_matrix;
    delete BBt;
}

void FastICA::run()
{
    #ifdef USE_TAU_PROFILING
    TAU_PROFILE( "FastICA::run()", "void()", TAU_DEFAULT );
    #endif
    double alpha = 1;
    double beta = 0;

    search_for_weights();
}

void FastICA::search_for_weights()
{
    #ifdef USE_TAU_PROFILING
    TAU_PROFILE( "FastICA::search_for_weights()", "void()", TAU_DEFAULT );
    #endif

    bool done = false;
    //int max_retries = 1;
    int retries = 0;
    steps   = 0;

    for (int i = 0; i < n_channels; i++)
    {
        done = false;
        steps = 0;
        retries = 0;

// This used to be in the while loop. How do we want to handle restarts?
        #ifdef USE_TAU_PROFILING
        TAU_PROFILE( "outerloop", "", TAU_DEFAULT );
        #endif

        #ifdef __DISTRIBUTED 
        if (rank == 0) 
        {
            initialize_w(i);
        }

        MPI_Bcast(w->data, w->n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        #else
        initialize_w(i);
        #endif

        while (retries < max_retries && !done)
        {

            #ifdef USE_FASTICA_VERBOSE                
            {
                m_rb.entry();
                cout << "[" << rank << "] after initialize_w[ " << i + 1 
                    << " ] " << endl;
                cout << "[" << rank << "] " << *w << endl;
                m_rb.exit();
            }
            #endif

            w->orthogonalize(BBt);
            w->normalize();


            #ifdef USE_FASTICA_VERBOSE                
            {
                m_rb.entry();
                cout << "[" << rank << "] " 
                    << "after orthonormal..before convergence[ " 
                    << i + 1 << " ] " << endl;
                cout << "[" << rank << "] " << *w << endl;
                cout << endl;
                m_rb.exit();
            }
            #endif

            convergence(i);

            if (steps >= max_iterations && tolerance > 0)
            {
                ++retries;
                if (retries >= max_retries)
                {
                    cout << "FastICA failed to converge on weight " 
                      << i << endl;
                    #ifdef USE_TAU_PROFILING
                    TAU_REPORT_STATISTICS();
                    #endif
                    #ifdef __DISTRIBUTED
                    MPI_Finalize();
                    #endif

                    exit(0);
                }
                else
                {

                    steps = 0;
                    #ifdef __DISTRIBUTED 
                    if (rank == 0) 
                    {
                    
                    cout << "Failed to converge on weight " << i << ".";
                    cout << " Restarting with new weight vector." << endl;
                   
                        initialize_w(i);
                    }

                    MPI_Bcast(w->data, w->n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
                    #else
                  
                    cout << "Failed to converge on weight " << i << ".";
                    cout << " Restarting with new weight vector." << endl;
                    
                    initialize_w(i);
                    #endif

                }
            }
            else
            {
                weights->add_row(w, i); // do we want to add column?
                double alpha = 1;
                int incx = 1;
                retries = 0; // this was retries == 0 ?why? cch

                dger_(&n_channels, &n_channels, &alpha, w->data,
                      &incx, w->data, &incx, BBt->data, &n_channels);


                #ifdef USE_FASTICA_VERBOSE                
                {
                    m_rb.entry();
                    cout << "[" << rank << "] " << "Channel[ " << i + 1 
                        << " ]..."  << *w << endl;
                    m_rb.exit();
                }
                #endif
                done = true;
            }
        }

        if(monitor_conv && rank == 0)
        {
            cout << "Component " << i + 1 << ", error = " << error << endl;
        }
    }
}

void FastICA::initialize_w (int i)
{
    if (  m_useRandomWeights )
    {
        double sum = 0.0;
        for ( int j = 0; j < n_channels; ++j )
        {
            (w->data)[j] = m_randNumGen.rand_double_closed(-1, 1);
            sum += ((w->data)[j] * (w->data)[j]);
        }
        sum = sqrt(sum);
        for ( int j = 0; j < n_channels; ++j )
        {
            (w->data)[j] /= sum;
        }
    }
    else 
    {
        memcpy(w->data,init_matrix->data + (i * n_channels),
            n_channels * sizeof(double));
    }
}

void FastICA::convergence( int i )
{
    double * d = data->data;
    int      n_channels = data->rows;
    int      n_observations = data->columns;


    #ifdef __DISTRIBUTED
    double      temp_error = 100;
    #endif

    // cch begin added code
    // I added this code to match the matlab code line for line
    // I agree, it's redundant, but was necessary for validation
    error = 100;
    steps = 0;
    w->orthogonalize(BBt);  // Bud sez this is unnecessary cuz it's already done just before calling
    w->normalize();         // convergence(). Bud wonders whether this ortho/normalize pair should
                            // be moved to just before the contrast() call
    // cch end added code

    int iterations = 0;
    while(!(error < tolerance) && steps < max_iterations)
    {
        #ifdef USE_TAU_PROFILING
        TAU_PROFILE( "innerloop", "", TAU_DEFAULT );
        #endif

        #ifdef USE_FASTICA_VERBOSE               
        if ( steps < 5 )
        {
            m_rb.entry();
            cout << "[" << rank << "] " << "convergence[" << 
                steps << "]" << " before contrast... w=" << endl;
            cout << "[" << rank << "] " << *w << endl << endl;
            m_rb.exit();
        }
        #endif
        
        contrast(data, w, w_plus); 

        #ifdef __DISTRIBUTED
        w_temp->zero_vector();                                            
        MPI_Allreduce(w_plus->data, w_temp->data, n_channels, MPI_DOUBLE, 
                      MPI_SUM, MPI_COMM_WORLD);		    		          
        w_plus->swap_vectors(w_temp);		    		                
        #endif

        #ifdef USE_FASTICA_VERBOSE                
        if ( steps < 5 )
        {
            m_rb.entry();
            cout << "[" << rank << "] " << "convergence[" << steps << "]" 
                << " before orthonormal... w_plus=" << *w_plus << endl;
            m_rb.exit();
        }
        #endif

        w_plus->orthogonalize(BBt);
        w_plus->normalize();

        #ifdef USE_FASTICA_VERBOSE                
        if ( steps < 5 )
        {
            m_rb.entry();
            cout << "[" << rank << "] " << "convergence[" << steps << "]" 
                << " after orthonormal... w_plus=" << *w_plus << endl;
            m_rb.exit();
        }
        #endif
        
        error = estimate_convergence(w, w_plus); 
        w->swap_vectors(w_plus);
        w_plus->zero_vector();

        ++steps;
    } // end of search for ith component

    if(monitor_conv) 
    {
        #ifdef __DISTRIBUTED
        if ( rank == 0 )
        #endif
        {
            cout << "[ " << rank << " ] iter:" 
                << ( steps + 1 ) << " W=" << *w << endl;
        }
    }
}

double FastICA::estimate_convergence(nic_vector* W,
                                     nic_vector* W_plus)
{
    double * w = W->data;
    double * w_plus = W_plus->data;
    int n_channels = W->n;

    double error = 0, err = 0;
    double error_hi = 0, error_lo = 0;

    for (int i = 0; i < n_channels; ++i)
    {
        error_hi += (w[i] - w_plus[i]) * (w[i] - w_plus[i]);
        error_lo += (w[i] + w_plus[i]) * (w[i] + w_plus[i]);
    }

    error_hi = sqrt(error_hi);
    error_lo = sqrt(error_lo);
    error = error_hi < error_lo ? error_hi : error_lo;

    return  error;
}

void FastICA::load_init_matrix(
    FastICAReader * reader, 
    nic_matrix* sph)
{
    int initialization_type= reader->initialization_type;

    if (initialization_type == 'r' || initialization_type == 'R')
    {
        init_matrix->random_matrix();
    }
    else if (initialization_type == 'i' || initialization_type == 'I')
    {
        init_matrix->identity_matrix();
    }
    else if (initialization_type == 'u' || initialization_type == 'U')
    {
        reader->load_user_init( init_matrix );
    }
    else
    {
        // error
    }
    
    //
    //	Apply the sph to the initial weight matrix
    //
    
}

void FastICA::load_contrast(char contrast_type)
{
    if (contrast_type == 'c' || contrast_type == 'C')
    {
        contrast = &cubic;
    }
    else if (contrast_type == 'g' || contrast_type == 'G')
    {
        contrast = &gaussian;
    }
    else if (contrast_type == 't' || contrast_type == 'T')
    {
        contrast = &hyp_tan;
    }
    else
    {
        // error
    }
}

void FastICA::validate_load()
{
    cout << "FastICA LOAD VALIDATION --------------------------------" << endl;
    cout << "\ttolerance             = " << tolerance << endl;
    cout << "\tmax_iterations        = " << max_iterations << endl;
    cout << "\tmax_retries           = " << max_retries << endl;
    cout << "\tsteps                 = " << steps << endl;
    cout << "\terror                 = " << error << endl;
    cout << "\tn_channels            = " << n_channels << endl;
    cout << "\tn_observations        = " << n_observations << endl;
    cout << "\trank                  = " << rank << endl;
    cout << "\ttotal_size            = " << total_size << endl;

    cout << "DATA SETS: first five elements from first column" << endl;

    cout << "\tdata: ";
    for (int i = 0; i < 5; i++)
    {
        cout << data->data[i] << " ";
    }
    cout << endl;

    cout << "\tinit_matrix: ";
    for (int i = 0; i < 5; i++)
    {
        cout << init_matrix->data[i] << " ";
    }
    cout << endl;

    cout << "END PREPROCESSOR LOAD VALIDATION ----------------------------" << endl;
}

#endif
// FASTICA_CPP_UNIVERSITY_OF_OREGON_NIC
