#ifndef INFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC
#define INFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC

#include "Infomax.h"
#include <math.h>

double ONE_  = 1;
double ZERO_ = 0;
int ONE  = 1;
int ZERO = 0; 

bool rtn;


Infomax::Infomax(
    nic_matrix* d, 
    InfomaxReader * reader, 
    nic_matrix* sph)
:   x(d), 
    n_channels(d->rows), n_observations(d->columns), m_sph( sph )
{
    nic_matrix::incrementReferenceCount( d );

    extended = false;
    exceeded = false;

    allocate_space(reader);

    // configure neural network ----------------------------------
    if (reader->nn_random_learning == 'y' || 
        reader->nn_random_learning == 'Y') 
    {

        random_learning = true;
    } 
    else 
    {
        random_learning = false;
    }
    max_weight = reader->nn_max_weight;
    weight_restart_factor = reader->nn_weight_restart_factor;

    // configure convergence loop ----------------------------------
    max_retries = reader->max_retries;
    max_iterations = reader->max_iterations;
    tolerance = reader->tolerance;
    learning_rate = reader->nn_learning_rate;//*******/log((double)n_channels);
    annealing_degree = reader->nn_annealing_degree;
    annealing_scale = reader->nn_annealing_scale;
    max_divergence = reader->nn_max_divergence;
    divergence_factor = reader->nn_divergence_factor;
    min_learning_rate = reader->nn_min_learning_rate;
    initial_weights = NULL;
}

Infomax::~Infomax()
{
    delete bias_load;
    delete temp_v;
    delete x_training;
    delete weights;
    delete old_weights;
    delete weights_delta;
    delete old_weights_delta;
    delete u;
    delete y;
    delete bias;
    delete x_index;
    delete init_matrix;
    delete temp_m;
    nic_matrix::freeReference( x );
}

void Infomax::allocate_space( InfomaxReader* reader )
{
    block_size = (int)ceil(5*log((double)n_observations));
    block_size = block_size < (int)ceil(0.3*(double)n_observations) ?  block_size :
                                            (int)(0.3 * (double)n_observations);

    //  block_size=35;
    n_blocks = n_observations/block_size;

    bias_load         = new nic_matrix(n_channels, block_size);
    temp_m            = new nic_matrix(n_channels, n_channels);
    temp_v            = new nic_vector(n_channels);

    x_training        = new nic_matrix(n_channels, block_size);
    weights           = new nic_matrix(n_channels, n_channels);
    old_weights       = new nic_matrix(n_channels, n_channels);


    weights_delta     = new nic_matrix(n_channels, n_channels);
    old_weights_delta = new nic_matrix(n_channels, n_channels);

    u                 = new nic_matrix(n_channels, block_size);
    y                 = new nic_matrix(n_channels, block_size);

    bias              = new nic_vector(n_channels);
    x_index           = new IndexArray(n_observations, 123456UL);
    init_matrix       = new nic_matrix(n_channels, n_channels);

    load_init_matrix(reader, m_sph);
}

void Infomax::load_init_matrix(InfomaxReader * reader, nic_matrix* sph)
{
    char initialization_type = reader->initialization_type;

    if (initialization_type == 'r' || initialization_type == 'R') 
    {
        init_matrix->random_matrix();
    } 
    else if (initialization_type == 'i' || initialization_type == 'I') 
    {
        init_matrix->identity_matrix();
    } 
    else if (initialization_type == 'u' || initialization_type == 'U') 
    {
        reader->load_user_init(init_matrix);
    } 
    else 
    {
    }
    
    //
    //    TODO:Apply sph to the initial guess
    //
    
}

// CLEAN UP DELTA, STOP USING temp2 FOR CONVERGENCE ESTIMATION
void Infomax::init_nn()
{
    init_delta      = tolerance + 10;

    delta           = init_delta;
    delta_angle     = 0;
    old_delta       = 0;

    *weights        = *init_matrix; // copy init matrix to weight matrix
    *old_weights    = *init_matrix; // copy init matrix to old weight matrix

    weights_delta->zero_matrix();
    old_weights_delta->zero_matrix();
    bias->zero_vector();
}

void Infomax::run()
{
    #ifdef __DISTRIBUTED
    int worldSize;
    MPI_Comm_size( MPI_COMM_WORLD, &worldSize );
    if ( worldSize > 1 )
    {
        int worldRank;
        MPI_Comm_rank( MPI_COMM_WORLD, &worldRank );
        if ( worldRank == 0 )
        {
            std::cerr 
                << "Infomax will produce bogus results if run as an MPI Process\n"
                << "on more than one processor. The MPI_COMM_WORLD size is " 
                << worldSize << "\n"
                << "processors. Attempting to exit cleanly from Infomax::run().\n" 
                << "Infomax is parallelized using OpenMP. This version of hipersat\n"
                << "is "
                #ifndef __OPENMP_
                << "not OpenMP enabled." << std::endl;
                #else
                << "OpenMP enabled. The number of processors can be set using the\n" 
                << "OMP_NUM_THREADS environment variable" << std::endl;
                #endif
        }
        MPI_Finalize( );
        exit( 1 );
    }
    #endif

    #ifdef USE_TAU_PROFILING
    TAU_PROFILE( "Infomax::run()", "void()", TAU_DEFAULT );
    #endif
    if(monitor_conv) 
    {
        init_nn();
        validate_load();
    }

    if (extended) 
    {
        search_for_weights_extended();
    } 
    else 
    {
        search_for_weights();
    }
}

void Infomax::validate_load()
{
    cout << "SEARCHING FOR THE WEIGHT SET" << endl;

    cout << "n_channels            = " << n_channels             << endl;
    cout << "n_observations        = " << n_observations         << endl;
    cout << "steps                 = " << steps                  << endl;
    cout << "retries               = " << retries                << endl;
    cout << "max_retries           = " << max_retries            << endl;
    cout << "max_iterations        = " << max_iterations         << endl;
    cout << "tolerance             = " << tolerance              << endl;
    cout << "delta                 = " << delta                  << endl;
    cout << "old_delta             = " << old_delta              << endl;
    cout << "init_delta            = " << init_delta             << endl;
    cout << "delta_angle           = " << delta_angle            << endl;
    cout << "exceeded              = " << exceeded               << endl;
    cout << "block_size            = " << block_size             << endl;
    cout << "n_blocks              = " << n_blocks               << endl;
    cout << "learning_rate         = " << learning_rate          << endl;
    cout << "annealing_scale       = " << annealing_scale        << endl;
    cout << "annealing_degree      = " << annealing_degree       << endl;
    cout << "random_learning       = " << random_learning        << endl;
    cout << "max_divergence        = " << max_divergence         << endl;
    cout << "divergence_factor     = " << divergence_factor      << endl;
    cout << "max_weight            = " << max_weight             << endl;
    cout << "weight_restart_factor = " << weight_restart_factor  << endl;
    cout << "min_learning_rate     = " << min_learning_rate      << endl;
    cout << "extended              = " << extended               << endl;


    cout << "DATA SETS: first five elements from first column---------------" 
        << endl;

    cout << "\tx @ " << x << ": ";
    if (x != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << x->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tx_training @ " << x_training << ": ";
    if (x_training != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << x_training->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tinit_matrix @ " << init_matrix << ": ";
    if (init_matrix != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << init_matrix->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\told_weights @ " << old_weights << ": ";
    if (old_weights != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << old_weights->data[i] << " ";
        }
      cout << endl;
    }

    cout << "\tinitial_weights @ " << initial_weights << ": ";
    if (initial_weights != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << initial_weights->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tweights_delta @ " << weights_delta << ": ";
    if (weights_delta != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << weights_delta->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\told_weights_delta @ " << old_weights_delta << ": ";
    if (old_weights_delta != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << old_weights_delta->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tu @ " << u << ": ";
    if (u != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << u->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\ty @ " << y << ": ";
    if (y != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << y->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tbias @ " << bias << ": ";
    if (bias != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << bias->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\tbias_load @ " << bias_load << ": ";
    if (bias_load != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << bias_load->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\ttemp_m @ " << temp_m << ": ";
    if (temp_m != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << temp_m->data[i] << " ";
        }
        cout << endl;
    }

    cout << "\ttemp_v @ " << temp_v << ": ";
    if (temp_v != NULL) 
    {
        for (int i = 0; i < 5; i++) 
        {
            cout << temp_v->data[i] << " ";
        }
        cout << endl;
    }

    if(random_learning) 
    {
        cout << "Using random temporal permutations of data" << endl;
    } 
    else
    {
        cout << "Using temporally ordered data" << endl;
    }
}

void Infomax::search_for_weights_extended()
{
    if(monitor_conv) 
    {
        cout << "SEARCHING FOR THE WEIGHT SET -- USING EXTENDED ALGORITHM--" 
            << endl;
        cout << "\tthis is not yet implemented" << endl;
    }
}

void Infomax::search_for_weights()
{
    #ifdef USE_TAU_PROFILING
    TAU_PROFILE( "Infomax::search_for_weights()", "void ()", TAU_DEFAULT );
    #endif

    bool valid = true;

    int steps   = 0;
    int retries = 0;
    max_retries = 6;
    int pid = 0;
    int procs;

    init_nn();

    #ifdef __OPENMP_

    #pragma omp parallel private(pid, procs)  // start parallel block
    {

    #endif

    #ifdef __OPENMP_
    #pragma omp critical
        {
            cerr << "n_procs=" << omp_get_num_threads() << endl;
            cerr << "max_procs=" << omp_get_max_threads() << endl;
            cerr << "which_procs=" << omp_get_thread_num() << endl;
        }
    #endif

        cout << "maximum iterations: " << max_iterations << endl;
        while(steps < max_iterations && 
            delta > tolerance && 
            retries < max_retries) 
        {

        #ifdef USE_TAU_PROFILING
        TAU_PROFILE( "innerloop", "", TAU_DEFAULT );
        #endif

        // Load index set, for random learing, permute index set order
        #ifdef __OPENMP_
        #pragma omp master  // master node only ##################################
        {
        #endif
            if (random_learning) 
            {
                x_index->permute_uniform();
            } 
            else 
            {
                x_index->ordered_indices();
            }
   
        #ifdef __OPENMP_
        }
        #pragma omp barrier
        #endif

        #ifdef USE_INFOMAX_DEBUG
        double * Wtemp  = weights->data;
        //double * Woldtemp = old_weights->data;
        int sizetemp = weights->size;
        cout<<"****************\n";
        for(int i=0;i<sizetemp;i++) 
        {
            cout<<Wtemp[i]<<" ";
            if((i+1)%4 ==0 )  cout<<"\n";
        }
        #endif
        // estimate the weight set using NN, then check the delta

        valid = training_loop(pid);
    
        #ifdef __OPENMP_
        #pragma omp master  // master node only ##################################
        {
        #endif
            if (valid) 
            {
                delta = estimate_convergence(pid);
                ++steps;
                if (steps > 2) 
                {
                    if (monitor_conv) 
                    {
                        cout << steps << ": learning rate = " << learning_rate;
                        cout << ", delta = " << delta << endl;
                    }
                    anneal_learning_rate();

                    if (delta > max_divergence) 
                    {
                        learning_rate *= divergence_factor;
                    }

                } 
                else if (steps == 1) 
                {
                    old_delta = delta;
                    *old_weights_delta = *weights_delta;
                    if (monitor_conv) 
                    {
                        cout << steps << ": learning rate = " 
                            << learning_rate;
                        cout << ", delta = " << delta << endl;
                    }
                } 
                else 
                {
                    if(monitor_conv) 
                    {
                        cout << steps << ": learning rate = " 
                            << learning_rate;
                        cout << ", delta = " << delta << endl;
                    }
                }
            } 
            else 
            {
                steps = 0;
                learning_rate *= weight_restart_factor;

                if (monitor_conv) 
                {
                    cout << "restart with new learning rate: " 
                        << learning_rate << endl;
                }

                if (learning_rate < min_learning_rate) 
                {
                    cout << "FAILURE: Infomax could not converge on a solution" << endl;
                    exit(0);
                }

                init_nn();
            }
            #ifdef __OPENMP_
        } 
        #pragma omp barrier
        #endif
        }  // end of search loop

    #ifdef __OPENMP_
    } // end of parallel block
    #endif
}//

double Infomax::estimate_convergence( int pid )
{
    double * W  = weights->data;
    double * delta_W = weights_delta->data;
    double * W_ = old_weights->data;
    int size = weights->size;


    //  Dw = weights - W_;
    for (int i = 0; i < size; i++) {
      *delta_W++ = *W++ - *W_++;
    };
    *old_weights = *weights;

    double delta_ = 0;
    delta_W = weights_delta->data;

    // delta = sum (Dw * Dw)
    for (int i = 0; i < size; i++) {
      delta_ += delta_W[i] * delta_W[i];
    };


    return delta_;
}

void Infomax::anneal_learning_rate()
{
    double * W = weights_delta->data;
    double * W_ = old_weights_delta->data;
    double diff = 0;
    int size = weights_delta->size;
    int i;

    // compute dw*dw'
    // see about a reduce here
    for (i = 0; i < size; i++) 
    {
        diff += W[i] * W_[i];
    }

    delta_angle = acos(diff/sqrt(delta*old_delta)) * 180./3.1415926;

    if (monitor_conv) 
    {
        cout << "DELTA ANGLE (deg)  = " << delta_angle << endl;
    }

    if (delta_angle > annealing_degree) 
    {
        learning_rate *= annealing_scale;
        old_delta = delta;
        // store current weights in old weights
        *old_weights_delta = *weights_delta;
     }
}

// Neural network training loop
bool Infomax::training_loop( int pid )
{
    int size = u->size;
    double * y_ = y->data;
    rtn = false; // global
    
    #ifdef USE_INFOMAX_DEBUG
    cout<<"\nWeights at the beginning of this iteration\n";
    for(int i=0;i<16;i++) 
    {
        cout<<weights->data[i]<<" ";
        if((i+1)%4==0) cout<<"\n";
    }
    #endif

    for (int i = 0; i < n_blocks && !rtn; i++) 
    {
        set_u(i, pid);
        set_y(pid);
        update_W(size, pid);

        #ifdef __OPENMP_
        #pragma omp master
        {
        #endif

            update_bias(pid);
            rtn = weight_exceeded(pid, i);

        #ifdef __OPENMP_
        }
        #pragma omp barrier
        #endif
    }
    return !rtn; // the calling functions is looking for valid completion
}

//Compute 
//u = Wx + bias
//first compute u = W*x, then compute u += bias
void Infomax::set_u(int idx, int pid) 
{
    double * u_;
    double * b  = bias->data;
    double * x_t;
    int    * index;

    int i, j;
    u_ = u->data;

    index = x_index->index + idx * block_size;


    #ifdef __OPENMP_
    #pragma omp for private(i, j) schedule(static)
    #endif
    for (i = 0; i < block_size; i++) 
    {  
        // set u = W*x using DGEMV (see below for details)
        u_ = u->data + i * n_channels;  // store results in next u vector
        x_t = x->data + index[i] * n_channels;  // store results in next u vecto

        #if defined( __OPENMP_ ) && defined( DGEMV_IS_NOT_THREADSAFE )
        #pragma omp critical 
        { 
            dgemv_("N", &n_channels, &n_channels, &ONE_, weights->data, 
                &n_channels, x_t, &ONE, &ZERO_, u_, &ONE);
        }
        #else
        dgemv_("N", &n_channels, &n_channels, &ONE_, weights->data, 
            &n_channels, x_t, &ONE, &ZERO_, u_, &ONE);
        #endif
   
        for (j = 0; j < n_channels; j++) 
        {
            u_[j] += b[j];
        }
    }

    #ifdef USE_INFOMAX_DEBUG
    cout<<"\nUUUUUUU\n";
    for(int i=0;i < u->size;i++) 
    {
        cout<<u->data[i]<<" ";
        if( (i+1)%4 == 0) cout<<'\n';
    }
    cout<<'\n';    
    #endif
}

// apply the activation function, we may want to make this function 
// more general
// yij = 1./(1.+exp(-uij));
void Infomax::set_y(int pid) 
{
    double * y_;
    double * u_;
    int size = y->size;
    int i;

    y_ = y->data;
    u_ = u->data;
    #ifdef __OPENMP_
    #pragma omp for private(i) schedule(static)
    #endif
    for (i = 0; i < size; i++) 
    {
        y_[i] = 1./(1. + exp(-u_[i]));
    }
    #ifdef USE_INFOMAX_DEBUG
    cout<<"\nYYYYYY\n";
    for(int i=0;i < y->size;i++) 
    {
        cout<<y_[i]<<" ";
        if((i+1)%4 == 0) cout<<'\n';
    }
    cout<<'\n';
    #endif
}

/*  
  w = w + dw = w + lr*(b*I + (1-2y)u')*w, 
  where w = weights, b = block_size, lr = learning_rate     
*/
void Infomax::update_W(int size, int pid) 
{

    double * y_;
    double * u_  = u->data;
    double * bl;
    double * tm;
    double * w;
    double * Dw;
    int i;
    weights_delta->zero_matrix();

    bl   = bias_load->data;
    y_  = y->data;
    tm  = temp_m->data;
    w   = weights->data;
    Dw  = weights_delta->data;


    #ifdef __OPENMP_
    #pragma omp for private(i) schedule(static)
    #endif
    for (i = 0; i < size; i++) 
    {
        bl[i] = 1 - 2*(y_[i]);
    }

    #ifdef CHOGE_DGEMM
    {
        int nthreads, tid, blSize, extraRows, totalSize;
        tid = omp_get_thread_num();
        nthreads = omp_get_num_threads();
        blSize = n_channels / nthreads;
        if ( tid == (nthreads - 1) )
        {
            extraRows = n_channels % nthreads;
        }
        else
        {
            extraRows = 0;
        }
        totalSize = blSize + extraRows;
        dgemm_("N", "T",  // transpose flags
               &totalSize, &n_channels, &block_size, // matrix sizes
               &ONE_, 
               bias_load->data + (tid * (totalSize -extraRows)) , &n_channels,
               u_, &n_channels,
               &ZERO_,
               tm + (tid * (totalSize-extraRows)), &n_channels );
        #pragma omp barrier
    }
    #else

    #if defined( USE_DGEMM_KGLASS )
    ABT(bias_load, u, temp_m);
    #elif defined( USE_DGEMM_MASTER )
    #ifdef __OPENMP_
    #pragma omp master  // master node only ##################################
    {
    #endif
        dgemm_("N", "T", &n_channels, &n_channels, &block_size,
            &ONE_, bias_load->data, &n_channels, u_, &n_channels, 
            &ZERO_, tm, &n_channels);
    #ifdef __OPENMP_
    } 
    #pragma omp barrier
    #endif
    #else
    dgemm_("N", "T", &n_channels, &n_channels, &block_size,
        &ONE_, bias_load->data, &n_channels, u_, &n_channels, &ZERO_, 
        tm, &n_channels);
    #endif

    #endif 
    // CHOGE_DGEMM

    #ifdef __OPENMP_
    #pragma omp for private(i) schedule(static)
    #endif
    for (i = 0; i < n_channels; i++) 
    {
        tm[i * n_channels + i] += block_size;
    }

    #ifdef CHOGE_DGEMM
    {
        int nthreads, tid, blSize, extraRows, totalSize;
        tid = omp_get_thread_num();
        nthreads = omp_get_num_threads();
        blSize = n_channels / nthreads;
        if ( tid == (nthreads - 1) )
        {
            extraRows = n_channels % nthreads;
        }
        else
        {
        extraRows = 0;
        }
        totalSize = blSize + extraRows;
        dgemm_("N", "N",  // transpose flags
            &totalSize, &n_channels, &n_channels, // matrix sizes
            &learning_rate, 
            tm + (tid * (totalSize -extraRows)) , &n_channels,
            w, &n_channels,
            &ZERO_,
            weights_delta->data + (tid * (totalSize-extraRows)), &n_channels );
        Dw = weights_delta->data;
        #pragma omp barrier
    }

    #else
    #if defined( USE_DGEMM_KGLASS )
    xAB(temp_m, weights, weights_delta, learning_rate);
    Dw = weights_delta->data;
    #elif defined( USE_DGEMM_MASTER )
    #ifdef __OPENMP_
    #pragma omp master  // master node only ##################################
    {
    #endif
        dgemm_("N", "N", &n_channels, &n_channels, &n_channels,
          &learning_rate, tm, &n_channels, w, &n_channels, 
          &ZERO_, weights_delta->data, &n_channels);
        Dw = weights_delta->data;
    #ifdef __OPENMP_
    }                 // ---------------------------------------------------
    #pragma omp barrier
    #endif
    #else
    dgemm_("N", "N", &n_channels, &n_channels, &n_channels,
        &learning_rate, tm, &n_channels, w, &n_channels, 
        &ZERO_, weights_delta->data, &n_channels);
    Dw = weights_delta->data;
    #endif

    #endif
    // CHOGE_DGEMM

    #ifdef __OPENMP_
    #pragma omp for private(i) schedule(static)
    #endif
    for (i = 0; i < weights->size; i++) 
    {
        w[i] += Dw[i];
    }
}

// bias = learning_rate * bsum + bias
void Infomax::update_bias(int pid)
{
    double * b;
    double * t;
    double * tv;
    int i, j;

    temp_v->zero_vector();

    b = bias->data;
    t = bias_load->data;
    tv = temp_v->data;

    // try to reduce rows of t, unlikely though
    for (i = 0; i < block_size; i++) 
    {
        t = bias_load->data + i * n_channels;
        for (j = 0; j < n_channels; j++) 
        {
            tv[j] += t[j];
        }
    }

    for (i = 0; i < n_channels; i++) 
    {
      *b++ += *tv++ * learning_rate;
    }

}

bool Infomax::weight_exceeded(int pid, int it) 
{
    bool rtn = false;
    double * w = weights->data;
    int      size = weights->size;
    double max = 0.;
    int i; 

    for (i = 0; i < size; i++) 
    {
        if (*w > max) 
        {
            max = *w;
        }
        ++w;
    }

    if (max > max_weight) 
    {
        rtn = true;
    } 

    return rtn;
}

void Infomax::ABT(nic_matrix* A, nic_matrix* B, 
    nic_matrix* C)
{
    int rows = A->rows;
    int cols = B->rows; // remember that B is transposed
    int p = A->columns;
    double *a_ = A->data;
    double *b_ = B->data;
    double *c_ = C->data;
    double *a_ptr = NULL;
    double *b_ptr = NULL;
    double *c_ptr = NULL;
    int i, j, k;

    //we could verify that B->cols = p, C->rows = rows and C->cols = cols
    #ifdef __OPENMP_
    #pragma omp for private(i, j, k, a_ptr, b_ptr, c_ptr)  schedule(static)
    #endif
    for (i = 0; i < rows; i++) 
    {
        for(j = 0; j < cols; j++) 
        {
            a_ptr = a_ + i;
            b_ptr = b_ + j;
            c_ptr = c_ + i + j * rows;
            *c_ptr = 0;
            for (k = 0; k < p; k++) 
            {
                *c_ptr += *a_ptr * *b_ptr;
                a_ptr += rows;
                b_ptr += rows;
            }
        }
    }
}

void Infomax::xAB(nic_matrix* A, nic_matrix* B, 
    nic_matrix* C, double x)
{

    int rows = A->rows;
    int cols = B->columns;
    int p = A->columns;
    double *a_ = A->data;
    double *b_ = B->data;
    double *c_ = C->data;
    double *a_ptr = NULL;
    double *b_ptr = NULL;
    double *c_ptr = NULL;
    int i, j, k;

    //we could verify that B->row = p, C->rows = rows and C->cols = cols
    #ifdef __OPENMP_
    #pragma omp for private(i, j, k, a_ptr, b_ptr, c_ptr) schedule(static)
    #endif
    for (int i = 0; i < rows; i++) 
    {
        for(int j = 0; j < cols; j++) 
        {
            a_ptr = a_ + i;
            b_ptr = b_ + j * p;
            c_ptr = c_ + i + j * rows;
            *c_ptr = 0;

            for (int k = 0; k < p; k++) 
            {
                *c_ptr += *a_ptr * *b_ptr;
                a_ptr += rows;
                ++b_ptr;
            }
            *c_ptr *= x;
        }
    }
}

#endif
// INFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC
