#ifndef MATRIXOPERATIONS_CPP_UNIVERSITY_OF_OREGON_NIC
#define MATRIXOPERATIONS_CPP_UNIVERSITY_OF_OREGON_NIC

#include "MatrixOperations.h"
#ifdef __OPENMP_
#include "omp.h"
#endif

void transposeSquare( nic_matrix& matrix )
{
    double temp;
    for ( int i = 0; i < matrix.rows; ++i )
    {
        for ( int j = 0; j < i; ++j )
        {
            temp = matrix( i, j );
            matrix( i, j ) = matrix( j, i );
            matrix( j, i ) = temp;
        }
    }
}

void parallelDgemm( 
    char* tA, char* tB,
    int* m, int* n, int* k,
    double* alpha, double* A, int* ldA,
    double* B, int* ldB,
    double* beta, double* C, int* ldC )
{
    bool transposeA = ( *tA == 'T' || *tA == 't' );
    int mTemp = *m;
#ifdef __PPC64_ESSL
    // empty
#elif defined( __OPENMP_ )
    int M = *m;
    int numThreads, threadNum;
#pragma omp parallel default(shared) firstprivate(mTemp, A, C, numThreads, threadNum, ldA )
    {
        numThreads = omp_get_num_threads();
        threadNum = omp_get_thread_num();
        mTemp = M / numThreads;
        A += ( transposeA ? (mTemp * threadNum * (*m)) : (mTemp * threadNum ));
        C += ( mTemp * threadNum );
        if ( threadNum == (numThreads-1) )
        {
            mTemp += M % numThreads;
        }
#endif
        dgemm_( tA, tB,
            &mTemp, n, k,
            alpha,
            A, ldA,
            B, ldB,
            beta, C, ldC );

#ifdef __PPC64_ESSL
#elif defined( __OPENMP_ )
    }
#endif
}

void parallelDgemv(
    char* tA,
    int* m, int* n,
    double* alpha, double* A, int* ldA,
    double* x, int* incx,
    double* beta, double* y, int* incy )
{
    bool transposeA = ( *tA == 'T' || *tA == 't' );
    int rows = ( transposeA ? *m : *n );
    int columns = ( transposeA ? *n : *m );
    int mTemp = *m;

#ifdef __PPC64_ESSL
#elif defined( __OPENMP_ )
    int M = *m;
    int numThreads, threadNum;
#pragma omp parallel default(shared) firstprivate( mTemp, A, y, numThreads, threadNum, ldA )
    {
        numThreads = omp_get_num_threads();
        threadNum = omp_get_thread_num();
        mTemp = M / numThreads;
        A += (transposeA ? (mTemp * threadNum * rows) : (mTemp * threadNum) );
        y += ( mTemp * threadNum );
        if ( threadNum == (numThreads - 1 ) )
        {
            mTemp += M % numThreads;
        }
#endif
        dgemv_( tA, &mTemp, n, alpha, A, ldA, x, incx, beta, y, incy );
#ifdef __PPC64_ESSL
#elif defined( __OPENMP_ )
    }
#endif
}


void multiply(
    nic_matrix& A,
    nic_matrix& B,
    nic_matrix& C,
    bool transposeA,
    bool transposeB )
{
    char tA = transposeA ? 'T' : 'N';
    char tB = transposeB ? 'T' : 'N';

    double alpha = 1;
    double beta = 0;
    int m = transposeA ? A.columns : A.rows;
    int n = transposeB ? B.rows :    B.columns;
    int k = transposeA ? A.rows :    A.columns;
    
    int lda = A.rows;
    int ldb = transposeB ? n : k;
    int ldc = m;

    double* aData = A.data;
    double* bData = B.data;
    double* cData = C.data;

    parallelDgemm( &tA, &tB, 
        &m, &n, &k, 
        &alpha, aData, &lda, 
        bData, &ldb, 
        &beta, cData, &ldc );
}

// in place multiplication, B is overwritten
// A can be transposed, B can not
void multiply(
    nic_matrix& A,
    nic_matrix& B,
    bool transposeA )
{
    double* temp = new double[ B.columns ];

    char tA = transposeA ? 'T' : 'N';
    double alpha = 1;
    double beta = 0;
    int m = transposeA ? A.columns : A.rows;
    int n = transposeA ? A.rows : A.columns;
    int lda = A.rows;
    int incx = 1;
    double* aData = A.data;
    double* bData = temp;
    double* cData;

    size_t columnLength = B.rows * sizeof( double );
    for ( int i = 0; i < B.columns; ++i )
    {
        cData = B.data + ( i * B.rows );
        memcpy( bData, cData, columnLength );
        dgemv_( &tA, &m, &n, &alpha, aData, &lda, bData, &incx,
            &beta, cData, &incx );
    }
    delete[] temp;
}

void multiply(
    nic_matrix& A,
    nic_vector& b,
    nic_vector& c,
    bool transposeA )
{
    char tA = transposeA ? 'T' : 'N';
    double alpha = 1;
    double beta = 0;
    int m = transposeA ? A.columns : A.rows;
    int n = transposeA ? A.rows : A.columns;
    int lda = A.rows;
    double* aData = A.data;
    double* bData = b.data;
    double* cData = c.data;
    int incx = 1;

    parallelDgemv( &tA, &m, &n, &alpha, aData, &lda, bData, &incx,
        &beta, cData, &incx );
}

void multiplyAdiag(
    nic_vector& A,
    nic_matrix& B,
    nic_matrix& C,
    bool transposeB )
{
    // for now we use our own hand-rolled implementation
    // TODO: investigate speeding up and parallelizing this

    int row = 0;
    int column = 0;
    int rows    = transposeB ? B.columns : B.rows;
    int columns = transposeB ? B.rows    : B.columns;
    int* rowPtr = transposeB ?    &column : &row;
    int* columnPtr = transposeB ? &row    : &column;

    for ( row = 0; row < rows; ++row )
    {
        for ( column = 0; column < columns; ++column )
        {
            double valB = B( *rowPtr, *columnPtr );
            double valA = A.operator()(row);
            C( row, column ) = valA * valB;
        }
    }
}

// slow matrix multiplication, used to verify that
// the OpenMP multiply routine works
void multiplySimple(
    nic_matrix& A,
    nic_matrix& B,
    nic_matrix& C,
    bool transposeA,
    bool transposeB )
{
    int aRows = transposeA ? A.columns : A.rows;
    int aColumns = transposeA ? A.rows : A.columns;
    int bRows = transposeB ? B.columns : B.rows;
    int bColumns = transposeB ? B.rows : B.columns;

    for ( int aRow = 0; aRow < aRows; ++aRow )
    {
        for ( int bColumn = 0; bColumn < bColumns; ++bColumn )
        {
            double sum = 0;
            for ( int aColumn = 0; aColumn < aColumns; ++aColumn )
            {
                // aColumn == bRow
                double valA = transposeA ? A( aColumn, aRow ) : A(aRow, aColumn );
                double valB = transposeB ? B( bColumn, aColumn ):  B(aColumn, bColumn );
                sum += valA * valB;
            }
            C( aRow, bColumn ) = sum;
        }
    }
}

bool invert( nic_matrix& A, nic_matrix& Ainv )
{
    Ainv = A;
    CLAPACK_integer rows = A.rows;
    CLAPACK_integer* ipiv = new CLAPACK_integer[ rows ];
    CLAPACK_integer info;
    double worksize;
    // compute the LU factorization
    dgetrf_( &rows, &rows, Ainv.data, &rows, ipiv, &info );
    bool success = false;
    if ( 0 == info )
    {
        CLAPACK_integer lwork = -1;
        dgetri_( &rows, Ainv.data, &rows, ipiv, &worksize, &lwork, &info );
        if ( 0 == info ) // inverse exists
        {
            lwork = (CLAPACK_integer)(worksize);
            double* workVector = new double[ lwork ];
            dgetri_( &rows, Ainv.data, &rows, ipiv, workVector, &lwork, &info );
            success = true;
            delete[] workVector;
        }
    }
    delete[] ipiv;
    return success;
}

#endif
// MATRIXOPERATIONS_CPP_UNIVERSITY_OF_OREGON_NIC
