#ifndef NICFASTICA_CPP_UNIVERSITY_OF_OREGON_NIC
#define NICFASTICA_CPP_UNIVERSITY_OF_OREGON_NIC

#include "NicFastICA.h"
#include "covariance.h"
#include "sphering_matrix.h"
#include "center_data.h"
#include "FastICANew.h"
#include "MatrixReader.h"
#include "MatrixWriter.h"
#include "MPIWrapper.h"
#include "nic_matrix.h"
#include "ContrastFunctions.h"
#include "MatrixOperations.h"
#include "average.h"
#include "PartitionedDataset.h"
#include "version.h"

#include <string>

using namespace std;


int main( int argc, char** argv )
{
    int rank, size;
    initializeMPI( argc, argv, rank, size );
    SimpleCLParser parser( argc, argv );

    setupParser( parser, rank );
    if ( rank == 0 )
    {
        cout << parser.usageHeader() << endl;
    }

    int rows = parser.readIntOption( "-c" );
    int columns = parser.readIntOption( "-s" );

    // determing the file formats
    MatrixReader::Format inputDataFormat = getInputFormat( parser );
    MatrixWriter::Format outputDataFormat = getOutputFormat( parser );

    nic_matrix inputData; 
    nic_matrix* spheringMatrix = 0;
    nic_vector avg( rows );
    if ( rank == 0 )
    {
        // read the data
        MatrixReader::loadMatrix(
            inputData,
            inputDataFormat,
            rows, columns,
            parser.readStringOption( "-i" ) );

        // center the data
        average( &inputData, &avg );
        center_data( &inputData, &avg );

        // compute the sphering matrix
        if ( parser.readFlagOption( "-sphering" ) ) // compute the sphering matrix
        {
            spheringMatrix = new nic_matrix( rows, rows );
            computeSpheringMatrix( &inputData, spheringMatrix, false );

            multiply( *spheringMatrix, inputData ); // in memory multiplication, overwrited inputData
        }
        else if ( parser.readFlagOption( "-is" ) )
        {
            spheringMatrix = new nic_matrix( rows, rows );
            MatrixReader::loadMatrix(
                *spheringMatrix,
                inputDataFormat,
                rows, rows,
                parser.readStringOption( "-is" ) );
        }
    }

    PartitionedDataset::distributeMatrix( &inputData );
    PartitionedDataset::broadcastMatrix( spheringMatrix );

    // FastICA
    FastICASettings settings;
    loadSettings( parser, settings );
    bool verbose = false;
    if ( rank == 0 ) verbose = true;
    FastICANew fastIca( settings, &inputData, 123456UL, verbose );
    if ( spheringMatrix != 0 )
    {
        fastIca.setWhiteningMatrix( spheringMatrix );
    }
    fastIca.runFastICA();

    // write the data

    if ( parser.readFlagOption( "-o" ) )
    {
        PartitionedDataset::gatherMatrix( &inputData );
    }

    if ( rank == 0 )
    {
        nic_matrix unmixingMatrix;
        fastIca.mixingMatrix( unmixingMatrix );
        // mixing matrix
        if ( parser.readFlagOption( "-om" ) )
        {
            nic_matrix mixingMatrix;
            invert( unmixingMatrix, mixingMatrix );
            MatrixWriter::writeMatrix(
                mixingMatrix, outputDataFormat, parser.readStringOption( "-om" ) );
                
        }

        // unmixing matrix
        if ( parser.readFlagOption( "-ow") )
        {
            MatrixWriter::writeMatrix(
                unmixingMatrix, outputDataFormat, parser.readStringOption( "-ow" ) );
        }

        // sphering matrix
        if ( parser.readFlagOption( "-os" ) && spheringMatrix != 0 )
        {
            MatrixWriter::writeMatrix( 
                *spheringMatrix, outputDataFormat, parser.readStringOption( "-os" ) );
        }

        if ( parser.readFlagOption( "-og" ) )
        {
            nic_matrix weightMatrix;
            fastIca.mixingMatrix( weightMatrix, false );
            MatrixWriter::writeMatrix(
                weightMatrix, outputDataFormat, parser.readStringOption( "-og" ) );
        }
        
        // the separated data
        if ( parser.readFlagOption( "-o" ) )
        {
            nic_matrix mixingMatrix;
            nic_vector unmixedAverages( rows );
            fastIca.mixingMatrix( mixingMatrix, false );

            // input data is replaced with unmixed signal
            multiply( mixingMatrix, inputData, false );


            fastIca.mixingMatrix( mixingMatrix );
            // unmix the averages
            multiply( mixingMatrix, avg, unmixedAverages );

            for ( int i = 0; i < rows; ++i )
            {
                for ( int j = 0; j < columns; ++j )
                {
                    inputData( i, j ) += unmixedAverages( i );
                }
            }

            MatrixWriter::writeMatrix(
                inputData, outputDataFormat, parser.readStringOption( "-o" ) );
        }
    }

    finalizeMPI();
    return 0;
}

void setupParser( SimpleCLParser& parser, int rank )
{
    bool isRequired = true;
    bool notRequired = false;

    // usage and copyright info
    string usage = "\nHiPerSAT FastICA v. 1.0 (r";
    usage += hipersatVersionString();
    usage +=  ")\n\n";
    usage += NIC_COPYRIGHT;
    parser.addUsageHeader( usage );


    // input file options
    parser.addOption( SimpleCLParser::STRING, "-i", "input data file name", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-if", "input data format ( big, little, native, text )", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-c", "number of input channels", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-s", "number of samples per channel", isRequired );

    // output options
    parser.addOption( SimpleCLParser::STRING, "-o", "unmixed data file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-om", "mixing matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-ow", "unmixing matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-os", "sphering matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-og", "weight matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-of", "output data format ( big, little, native, text )", notRequired );

    // sphering options
    parser.addOption( SimpleCLParser::FLAG, "-sphering", "compute sphering matrix from input data", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-is", "load precomputed sphering matrix from file", notRequired );

    // FastICA options
    parser.addOption( SimpleCLParser::DOUBLE, "-t", "convergence tolerance", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-I", "maximum numer of iterations per channel", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-C", "contrast function ( cubic, hyptan, gaussian )", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-g", "weight matrix initialization type ( identity, random, user )", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-ig", "weight matrix initialization file name", notRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-r", "maximum number of retries (using random restarts)", isRequired );

    // usage
    parser.addOption( SimpleCLParser::FLAG, "-h", "usage information", notRequired );

    parser.parse();
    if ( parser.readFlagOption( "-h" ) )
    {
        if ( rank == 0 )
        {
            cout << parser.usage() << endl;
        }
        exit( 0 );
    }

    parser.checkRequired();
}

MatrixReader::Format getInputFormat( SimpleCLParser& parser )
{
    string formatString = parser.readStringOption( "-if" );
    MatrixReader::Format returnValue;
    if ( formatString == "big" )
    {
        returnValue = MatrixReader::BIG;
    }
    else if ( formatString == "little" )
    {
        returnValue = MatrixReader::LITTLE;
    }
    else if ( formatString == "native" )
    {
        returnValue = MatrixReader::NATIVE;
    }
    else if ( formatString == "text" )
    {
        returnValue = MatrixReader::TEXT;
    }
    else
    {
        cerr << "Invalid file type: " << formatString << endl;
        exit(1);
    }
    return returnValue;
}

MatrixWriter::Format getOutputFormat( SimpleCLParser& parser )
{
    MatrixWriter::Format format = MatrixWriter::TEXT;

    if ( parser.readFlagOption( "-of" ) )
    {
        string formatName = parser.readStringOption( "-of" );
        if ( formatName == "big" )
        {
            format = MatrixWriter::BIG;
        }
        else if ( formatName == "little" )
        {
            format = MatrixWriter::LITTLE;
        }
        else if ( formatName == "native" )
        {
            format = MatrixWriter::NATIVE;
        }
        else if ( formatName == "text" )
        {
            format = MatrixWriter::TEXT;
        }
        else 
        {
            cerr << "Invalid output type. Defaulting to text" << endl;
        }
    }
    return format;
}

void loadSettings( SimpleCLParser& parser, FastICASettings& settings)
{
    settings.convergenceTolerance = parser.readDoubleOption( "-t" );
    settings.maximumIterations = parser.readIntOption( "-I" );
    string cf = parser.readStringOption( "-C" );
    if ( cf == "cubic" )
    {
        settings.contrastFunction = ContrastFunctions::CUBIC;
    }
    else if ( cf == "hyptan" )
    {
        settings.contrastFunction = ContrastFunctions::HYPERBOLIC_TAN;
    }
    else if ( cf == "gaussian" )
    {
        settings.contrastFunction = ContrastFunctions::GAUSSIAN;
    }
    else
    {
        cerr << "unknown constrast function: " << cf << endl;
        cerr << "exiting" << endl;
        exit( 1 );
    }
    string initType = parser.readStringOption( "-g" );
    if ( initType == "identity" )
    {
        settings.initializationType = FastICASettings::IDENTITY;
    }
    else if ( initType == "random" )
    {
        settings.initializationType = FastICASettings::RANDOM;
    }
    else if ( initType == "user" )
    {
        settings.initializationType = FastICASettings::USER_SPECIFIED;
    }
    else
    {
        cerr << "unknown initialization type " << initType << endl;
        cerr << "exiting" << endl;
        exit( 1 );
    }
    settings.userInitializationFile = parser.readStringOption( "-ig" );
    // TODO make this smarter
    settings.userInitializationFileFormat = MatrixReader::BIG;
    settings.maximumRetries = parser.readIntOption( "-r" );
}

#endif
// NICFASTICA_CPP_UNIVERSITY_OF_OREGON_NIC
