#ifndef NICINFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC
#define NICINFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC


#include "NicInfomax.h"
#include "covariance.h"
#include "sphering_matrix.h"
#include "center_data.h"
#include "InfomaxNew.h"
#include "MatrixReader.h"
#include "MatrixWriter.h"
#include "MPIWrapper.h"
#include "nic_matrix.h"
#include "ContrastFunctions.h"
#include "MatrixOperations.h"
#include "average.h"
#include "version.h"
#include "SortComponents.h"

#include <string>

using namespace std;

int main( int argc, char** argv )
{
    int rank, size;
    SimpleCLParser parser( argc, argv );

    setupParser( parser, rank );
    cout << parser.usageHeader() << endl;

    int rows = parser.readIntOption( "-c" );
    int columns = parser.readIntOption( "-s" );

    // determing the file formats
    MatrixReader::Format inputDataFormat = getInputFormat( parser );
    MatrixWriter::Format outputDataFormat = getOutputFormat( parser );

    nic_matrix inputData; 
    nic_matrix* spheringMatrix = 0;
    nic_vector avg( rows );
    // read the data
    MatrixReader::loadMatrix(
        inputData,
        inputDataFormat,
        rows, columns,
        parser.readStringOption( "-i" ) );

    // center the data
    average( &inputData, &avg );
    center_data( &inputData, &avg );

    // compute the sphering matrix
    if ( parser.readFlagOption( "-sphering" ) ) // compute the sphering matrix
    {
        spheringMatrix = new nic_matrix( rows, rows );
        computeSpheringMatrix( &inputData, spheringMatrix, true );

        multiply( *spheringMatrix, inputData ); // in memory multiplication, overwrited inputData
    }
    else if ( parser.readFlagOption( "-is" ) )
    {
        spheringMatrix = new nic_matrix( rows, rows );
        MatrixReader::loadMatrix(
            *spheringMatrix,
            inputDataFormat,
            rows, rows,
            parser.readStringOption( "-is" ) );
    }

    // Infomax
    InfomaxSettings settings;
    loadSettings( parser, settings );
    bool verbose = false;
    if ( rank == 0 ) verbose = true;
    unsigned long seed = 123456UL;
    if ( parser.readFlagOption( "-seed" ) )
    {
        seed = parser.readIntOption( "-seed" );
    }
    InfomaxNew infomax( settings, &inputData, seed, verbose );
    if ( spheringMatrix != 0 )
    {
        infomax.setWhiteningMatrix( spheringMatrix );
    }
    infomax.runInfomax();
    nic_matrix mixingMatrix;

    infomax.posact( inputData, mixingMatrix );
    sortComponents( inputData, mixingMatrix, *spheringMatrix );

    // write the data

    nic_matrix unmixingMatrix = mixingMatrix;
    //infomax.mixingMatrix( unmixingMatrix );

    // unmixing matrix
    if ( parser.readFlagOption( "-om" ) )
    {
        nic_matrix mixingMatrix;
        invert( unmixingMatrix, mixingMatrix );
        MatrixWriter::writeMatrix(
            mixingMatrix, outputDataFormat, parser.readStringOption("-om"));
    }

    if ( parser.readFlagOption( "-ow" ) )
    {
        MatrixWriter::writeMatrix(
            unmixingMatrix, outputDataFormat, parser.readStringOption("-ow"));
    }

    // sphering matrix
    if ( parser.readFlagOption( "-os" ) && spheringMatrix != 0 )
    {
        MatrixWriter::writeMatrix( 
            *spheringMatrix, outputDataFormat, parser.readStringOption("-os"));
    }

    if ( parser.readFlagOption( "-og" ) )
    {
        nic_matrix weightMatrix( rows, rows );
        //infomax.mixingMatrix( weightMatrix, false );
        nic_matrix invSphering( rows, rows );
        if ( spheringMatrix != 0 )
        {
            invert( *spheringMatrix, invSphering );
            multiply( unmixingMatrix, invSphering, weightMatrix );
        }
        else
        {
            weightMatrix = unmixingMatrix;
        }

        MatrixWriter::writeMatrix(
            weightMatrix, outputDataFormat, parser.readStringOption("-og"));
    }
    
        // the separated data
        if ( parser.readFlagOption( "-o" ) )
        {
            nic_matrix mixingMatrix;
            nic_vector unmixedAverages( rows );
            infomax.mixingMatrix( mixingMatrix, false );

            // input data is replaced with unmixed signal
            multiply( mixingMatrix, inputData, false );


            infomax.mixingMatrix( mixingMatrix );
            // unmix the averages
            multiply( mixingMatrix, avg, unmixedAverages );

            for ( int i = 0; i < rows; ++i )
            {
                for ( int j = 0; j < columns; ++j )
                {
                    inputData( i, j ) += unmixedAverages( i );
                }
            }

            MatrixWriter::writeMatrix(
                inputData, outputDataFormat, parser.readStringOption( "-o" ) );
        }

    return 0;
}

void setupParser( SimpleCLParser& parser, int rank )
{
    bool isRequired = true;
    bool notRequired = false;

    // usage and copyright info
    string usage = "\nHiPerSAT Infomax v. 1.0 (r";
    usage += hipersatVersionString();
    usage +=  ")\n\n";
    usage += NIC_COPYRIGHT;
    parser.addUsageHeader( usage );


    // input file options
    parser.addOption( SimpleCLParser::STRING, "-i", "input data file name", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-if", "input data format ( big, little, native, text )", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-c", "number of input channels", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-s", "number of samples per channel", isRequired );

    parser.addOption( SimpleCLParser::STRING, "-is", "input sphering matrix" , notRequired );

    // output options
    parser.addOption( SimpleCLParser::STRING, "-o", "unmixed data file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-om", "mixing matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-ow", "unmixing matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-os", "sphering matrix file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-og", "weight matrix file name", notRequired );

    parser.addOption( SimpleCLParser::STRING, "-of", "output data format ( big, little, native, text )", notRequired );

    // infomax options
//    parser.addOption( SimpleCLParser::FLAG, "-extended", "use extended infomax", notRequired );
//    parser.addOption( SimpleCLParser::FLAG, "-pca", "principal component analysis", notRequired );
//    parser.addOption( SimpleCLParser::INT, "-ncomps", "number if ICA components to compute (default channels)", notRequired );
    parser.addOption( SimpleCLParser::FLAG, "-sphering", "flag sphering of data", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-g", "initial weight matrix (identity, random, user)", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-ig", "file name for user weights", notRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-lrate", "initial ICA learning rate << 1 (default heuristic)", notRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-block", "ICA block size (default heuristic)", notRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-anneal", "annealing constant (default 0.90)", notRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-annealdeg", "annealing degree weight change (default 70)", notRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-stop", "weight change tolerance to stop training (default for channels < 34, 1e-6, else 1e-7)", notRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-maxsteps", "maximum number of ICA training steps", notRequired );
    parser.addOption( SimpleCLParser::FLAG, "-verbose", "give ascii messages of progress", notRequired );

    parser.addOption( SimpleCLParser::INTEGER, "-seed", "seed for random number generator (default 123456)", notRequired );

/*    // sphering options
    parser.addOption( SimpleCLParser::FLAG, "-S", "compute sphering matrix from input data", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-Si", "load precomputed sphering matrix from file", notRequired );

    // Infomax options
    parser.addOption( SimpleCLParser::DOUBLE, "-t", "convergence tolerance", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-I", "maximum numer of iterations per channel", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-r", "maximum number of retries (using random restarts)", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-w", "weight matrix initialization type ( identity, random, user )", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-wi", "weight matrix initialization file name", notRequired );
    parser.addOption( SimpleCLParser::STRING, "-C", "contrast function ( cubic, hyptan, gaussian )", isRequired );

    // neural net options
    parser.addOption( SimpleCLParser::INTEGER, "-nnBS", "block size", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnLR", "learning rate", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnAD", "annealing degree", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnAS", "annealing scale", isRequired );
    parser.addOption( SimpleCLParser::FLAG, "-nnRL", "use random learning", notRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnMW", "maximum weight", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnD", "divergence", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnMD", "maximum divergence", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnWRF", "weight restart factor", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnDF", "divergence factor", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-nnMLR", "minimum learning rate", isRequired );
    */

    // usage
    parser.addOption( SimpleCLParser::FLAG, "-h", "usage information", notRequired );

    parser.parse();
    if ( parser.readFlagOption( "-h" ) )
    {
        cout << parser.usage() << endl;
        exit(0);
    }

    parser.checkRequired();
}

MatrixReader::Format getInputFormat( SimpleCLParser& parser )
{
    string formatString = parser.readStringOption( "-if" );
    MatrixReader::Format returnValue;
    if ( formatString == "big" )
    {
        returnValue = MatrixReader::BIG;
    }
    else if ( formatString == "little" )
    {
        returnValue = MatrixReader::LITTLE;
    }
    else if ( formatString == "native" )
    {
        returnValue = MatrixReader::NATIVE;
    }
    else if ( formatString == "text" )
    {
        returnValue = MatrixReader::TEXT;
    }
    else
    {
        cerr << "Invalid file type: " << formatString << endl;
        exit(1);
    }
    return returnValue;
}

MatrixWriter::Format getOutputFormat( SimpleCLParser& parser )
{
    MatrixWriter::Format format = MatrixWriter::TEXT;

    if ( parser.readFlagOption( "-of" ) )
    {
        string formatName = parser.readStringOption( "-of" );
        if ( formatName == "big" )
        {
            format = MatrixWriter::BIG;
        }
        else if ( formatName == "little" )
        {
            format = MatrixWriter::LITTLE;
        }
        else if ( formatName == "native" )
        {
            format = MatrixWriter::NATIVE;
        }
        else if ( formatName == "text" )
        {
            format = MatrixWriter::TEXT;
        }
        else 
        {
            cerr << "Invalid output type. Defaulting to text" << endl;
        }
    }
    return format;
}

void loadSettings( SimpleCLParser& parser, InfomaxSettings& settings)
{
    if ( parser.readFlagOption( "-g" ) )
    {
        string value = parser.readStringOption( "-g" );
        if ( value == "identity" )
        {
            settings.initializationType = InfomaxSettings::IDENTITY;
        }
        else if ( value == "random" )
        {
            settings.initializationType = InfomaxSettings::RANDOM;
        }
        else if ( value == "user" )
        {
            settings.initializationType = InfomaxSettings::USER_SPECIFIED;
            settings.userInitializationFile = parser.readStringOption( "-userWeights" );
        }
    }
    else
    {
        settings.initializationType = InfomaxSettings::IDENTITY;
    }

    if ( parser.readFlagOption( "-lrate" ) )
    {
        settings.nnLearningRate = parser.readDoubleOption( "-lrate" );
    }
    else
    {
        double channels = parser.readIntOption( "-c" );
        settings.nnLearningRate = 0.00065 / ( log( channels ) );
    }

    if ( parser.readFlagOption( "-block" ) )
    {
        settings.nnBlockSize = parser.readIntOption( "-block" );
    }
    else
    {
        double samples = parser.readIntOption( "-s" );
        settings.nnBlockSize = (int)(ceil( 5.0*log( samples ) ) );
    }

    if ( parser.readFlagOption( "-anneal" ) )
    {
        settings.nnAnnealingScale = parser.readDoubleOption( "-anneal" );
    }
    else
    {
        settings.nnAnnealingScale = 0.90;
    }

    if ( parser.readFlagOption( "-annealdeg" ) )
    {
        settings.nnAnnealingDegree = parser.readDoubleOption( "-annealdeg" );
    }
    else
    {
        settings.nnAnnealingDegree = 70.0;
    }

    if ( parser.readFlagOption( "-stop" ) )
    {
        settings.convergenceTolerance = parser.readDoubleOption( "-stop" );
    }
    else
    {
        int  channels = parser.readIntOption( "-c" );
        if ( channels < 24 ) settings.convergenceTolerance = 1e-6;
        else settings.convergenceTolerance = 1e-7;
    }

    if ( parser.readFlagOption( "-maxsteps" ) )
    {
        settings.maximumIterations = parser.readIntOption( "-maxsteps" );
    }
    else
    {
        settings.maximumIterations = 512;
    }

    settings.nnWeightRestartFactor = 0.9;
    settings.nnDivergenceFactor = 0.8;
    settings.nnMaxWeight = 1e8;
    settings.nnMinLearningRate = 0.000001;
    settings.nnMaxDivergence = 1000000000;
    settings.nnRandomLearning = true;
    
    settings.maximumRetries = 100;

    // TODO make this smarter
    settings.userInitializationFormat = MatrixReader::BIG;
}

#endif
// NICINFOMAX_CPP_UNIVERSITY_OF_OREGON_NIC
