#ifndef PROCESSREADER_CPP_UNIVERSITY_OF_OREGON_NIC
#define PROCESSREADER_CPP_UNIVERSITY_OF_OREGON_NIC

#include "ProcessReader.h"

ProcessReader::ProcessReader( int r, int np, std::istream& formStream )
:   rank(r),
    n_cpus(np),
    monitor_conv(false)
{
    #ifdef USE_TAU_PROFILING
    TAU_PROFILE( "data_input", "void (int, int, istream&)", TAU_DEFAULT );
    #endif

    data = NULL;
    preprocessor = NULL;
    process = NULL;
    output = NULL;

    // loads information regarding source matrix and signal analysis technique
    CleanerReader * clean = 
        (CleanerReader *)( new Parser<CleanerReader>( formStream ));

    CleanerReader::PROCESS_TYPE type = clean->process;

    // data is allocated by load_matrix, but is not owned by it
    data = clean->load_matrix(rank, n_cpus); 
    nic_matrix::incrementReferenceCount( data );

    #ifndef __DISTRIBUTED 
    n_cpus = clean->n_procs; // required for shared memory only 
    #endif 
    
    if (type == CleanerReader::FASTICA) 
    {
        // configure the workbench for FastICA
        PreprocessReader * preprocess_reader = 
            (PreprocessReader *)new Parser<PreprocessReader>( formStream );
        preprocessor = new Preprocessor(
            data, preprocess_reader, rank, n_cpus, false );

        FastICAReader * fica_reader = 
            (FastICAReader *)new Parser<FastICAReader>( formStream );
        process = new FastICA(data, fica_reader, preprocessor->sph);

        ICAOutputReader * output_reader = 
            (ICAOutputReader *)new Parser<ICAOutputReader>( formStream ); 
        output = new ICAOutput(preprocessor, process, output_reader, type);

        #ifdef __DISTRIBUTED 
        monitor_conv = output_reader->monitor_conv & (rank == 0);
        #endif 

        validate = output_reader->validate;
        // delete the readers

        delete preprocess_reader;
        delete fica_reader;
        delete output_reader;

    } 
    else if (type == CleanerReader::INFOMAX) 
    {
        PreprocessReader * preprocess_reader = 
            (PreprocessReader *)new Parser<PreprocessReader>( formStream );

        InfomaxReader * infomax_reader = 
            (InfomaxReader *)new Parser<InfomaxReader>( formStream );
        ICAOutputReader * output_reader = 
            (ICAOutputReader *)new Parser<ICAOutputReader>( formStream );

        preprocessor = new Preprocessor(
            data, preprocess_reader, rank, n_cpus, true );
        process = new Infomax(data, infomax_reader, preprocessor->sph);
        output = new ICAOutput(preprocessor, process, output_reader, type);

        monitor_conv = output_reader->monitor_conv;
        validate = output_reader->validate;

        // delete the readers
        delete preprocess_reader;
        delete infomax_reader;
        delete output_reader;

    } 
    else if (type == CleanerReader::PCA) 
    {
    } 
    else if (type == CleanerReader::NONE) 
    {
        PreprocessReader * preprocess_reader = 
            (PreprocessReader *)new Parser<PreprocessReader>( formStream );
        preprocessor = new Preprocessor(
            data, preprocess_reader, rank, n_cpus, false );
        process = NULL;
        ICAOutputReader * output_reader = 
            (ICAOutputReader *)new Parser<ICAOutputReader>( formStream );

        monitor_conv = output_reader->monitor_conv;
        validate = output_reader->validate;

        // delete the readers
        delete preprocess_reader;

    } else 
    {

        std::cout << "ProcessReader  CONFIGURATION ERROR --" << std::endl;
        std::cout << "in the Signal cleaning form, please specify an ";
        std::cout << "appropriate process.\nThe choices are: " << std::endl;
        std::cout << "\tFastICA" << std::endl;
        std::cout << "\tInfomax" << std::endl;
        std::cout << "\tPCA" << std::endl;
        std::cout << "\tNone" << std::endl;
        exit(-3);
    }

    delete clean;

} // end method

ProcessReader::~ProcessReader() 
{
    if ( data != 0 ) nic_matrix::freeReference( data );
}

void ProcessReader::validate_load() 
{
    cout << "ProcessReader validation ---------------------------------" << endl;
    cout << "\tProcessReader: rank           = " << rank << endl;
    cout << "\tProcessReader: n_cpus         = " << n_cpus << endl;
    cout << "\tProcessReader: preprocessor   @ " << preprocessor << endl;
    cout << "\tProcessReader: process        @ " << process << endl;
    cout << "\tProcessReader: output         @ " << output << endl;
    cout << "\tProcessReader: monitor_conv   = " << monitor_conv << endl;
    cout << "\tProcessReader: data           @ " << data << endl;
    cout << "DATA SETS: first five elements from first column" << endl;

    if (data == NULL) 
    {
      cout << "ERROR: data not loaded" << endl;
      exit(-4);
    }
    cout << "\t: ";
    for (int i = 0; i < 5; i++) 
    {
        cout << data->data[i] << " ";
    }
    cout << endl;
    cout << "END PROCESS_READER LOAD VALIDATION ----------------------------" << endl;

    preprocessor->validate_load();
    process->validate_load();
    output->validate_load();
}


#endif
// PROCESSREADER_CPP_UNIVERSITY_OF_OREGON_NIC
