#ifndef SOBI_CPP_UNIVERSITY_OF_OREGON_NIC
#define SOBI_CPP_UNIVERSITY_OF_OREGON_NIC

#include "Sobi.h"
#include "MatrixOperations.h"
#include "nic_vector.h"
#include <math.h>

// dataMatrix: m x n
// tau: vector of time lags, it is ok if this vector is of length 0
// W: output estimate of the inverse of the mixing matrix
void computeSobi( nic_matrix& dataMatrix, int* tau, int& tauLength, 
    nic_matrix& W )
{
    int rows = dataMatrix.rows;
    int columns = dataMatrix.columns;

    float tolerance = 1e-8;

    // create the tau vector
    if ( 0 == tauLength || tau == 0 )
    {
        tauLength = columns / 3;
        if ( (columns % 3) > 0 ) ++tauLength;
        if ( tauLength > 100 ) tauLength = 100;
        tau = new int[ tauLength ];
        for ( int i = 0; i < tauLength; ++i )
        {
            tau[ i ] = ( i+1 );
        }
    }

    nic_matrix covX( rows, tauLength * rows );

    for ( int i = 0; i < tauLength; ++i )
    {
        int first = i * rows;
        computeStdCov( dataMatrix, tau[i], covX, first );
    }

    // joint diagonalization
    computeRjd( covX, W );
}

void computeStdCov( nic_matrix& dataMatrix, int tau, 
    nic_matrix& covX, int first )
{
    // we will need to wrap the code up to dgemm to make it
    // parallel; out parallel matrix operations don't support
    // using lda, ldb, and ldc, although it is likely that
    // the dgemm operation below won't have any problems

    int m = dataMatrix.rows;
    int columns = dataMatrix.columns;
    int n = m;
    int k = columns - tau;
    double alpha = 1.0 / (double)(k);
    double beta = 0.0;
    double* A;
    double* B;
    double* C;
    A = dataMatrix.data + (tau * m);
    B = dataMatrix.data;
    C = covX.data + (first * m);
    char tA = 'N';
    char tB = 'T';
    dgemm_( &tA, &tB, &m, &m, &k, &alpha, A, &m, B, &m, &beta, C, &m );
    float norm = frobeniusNorm( C, m );
    for ( int i = 0; i < m * m; ++i )
    {
        (*C) *= norm;
        ++C;
    }
}

float frobeniusNorm( double* matrix, int size )
{
    float norm = 0.0;
    int length = size * size;
    for ( int i = 0; i < length; ++i )
    {
        norm += (*matrix) * (*matrix);
        ++matrix;
    }
    norm = sqrt( norm );
    return norm;
}

// A is a m x nm matrix
void computeRjd( nic_matrix& A, nic_matrix& W )
{
    float tolerance = 1e-8;

    int rows = A.rows;
    int columns = A.columns / A.rows;
    assert( ( rows * columns ) == A.columns );

    W.resize( rows, rows );
    W.identity_matrix();

    bool encore = true;
    // g is a temp variable
    // G is the givens rotation
    nic_matrix g( 2, columns );
    nic_matrix G( 2, 2 );

    // p and q Matrices
    nic_matrix Mp( rows, columns );
    nic_matrix Mq( rows, columns );

    // p and q Rows
    nic_vector Rp( A.columns );
    nic_vector Rq( A.columns );

    // wTemp
    nic_vector Wtemp( rows );
    
    float ton;
    float toff;
    float theta;
    float c,s;
    int iteration = 0;
    while ( encore )
    {
        ++iteration;
        encore = false;
        for ( int p = 0; p < rows-1; ++p ) // loop p from 1 to rows - 1
        {
            for ( int q = p+1; q < rows; ++q ) // loop q from p to rows
            {
                // compute the Givens rotation
                for ( int j = 0; j < columns; ++j )
                {
                    g( 0, j ) = A( p, p + (rows * j)) - A( q, q + (rows * j));
                    g( 1, j ) = A( p, q + (rows * j)) + A( q, p + (rows * j));
                }
                multiply( g, g, G, false, true );
                ton = G(0,0) - G(1,1);
                toff = G(0,1) + G(1,0);
                float dist = sqrt( (ton*ton) + (toff*toff) );
                theta = 0.5 * atan2( toff, ton  + dist );
                c = cos( theta );
                s = sin( theta );
                encore = encore || ( fabs( s ) > tolerance );

                // update A and W

                if ( fabs( s ) > tolerance )
                {

                    // initialize Mp and Mq
                    for ( int i = 0; i < rows; ++i )
                    {
                        for ( int j = 0; j < columns; ++j )
                        {
                            Mp( i, j ) = A( i, p + (rows * j));
                            Mq( i, j ) = A( i, q + (rows * j));
                        }
                    }

                    // first A update, this can be a point of optimization
                    for ( int i = 0; i < rows; ++i )
                    {
                        for ( int j = 0; j < columns; ++j )
                        {
                            A( i, p + (rows * j)) = c * Mp(i,j) + s * Mq(i,j);
                            A( i, q + (rows * j)) = c * Mq(i,j) - s * Mp(i,j);
                        }
                    }

                    // initialize Rp and Rq
                    for ( int i = 0; i < A.columns; ++i )
                    {
                        Rp(i) = A( p, i );
                        Rq(i) = A( q, i );
                    }

                    // second A update
                    for ( int i = 0; i < A.columns; ++i )
                    {
                        A(p,i) = c * Rp(i) + s * Rq(i);
                        A(q,i) = c * Rq(i) - s * Rp(i);
                    }

                    // initialize Wtemp
                    for ( int i = 0; i < rows; ++i )
                    {
                        Wtemp(i) = W(p, i);
                    }

                     // update W
                     for ( int i = 0; i < rows; ++i )
                     {
                         W(p,i) = c*W(p,i) + s*W(q,i);
                         W(q,i) = c*W(q,i) - s*Wtemp(i);
                     }
                }
            }
        }
    }
}

#endif
// SOBI_CPP_UNIVERSITY_OF_OREGON_NIC
