#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <time.h>
#include <unistd.h> 

#include "ProcessReader.h"
#include "Preprocessor.h"
#include "Processor.h"
#include "nic_matrix.h"

#include "ICAOutput.h"

#include "MPIWrapper.h"

#ifdef __PERFORMANCE_TIMING
#include <sys/times.h> 
#include <sys/types.h>
#include <stropts.h> 
#include <poll.h> 
#endif

#ifdef USE_TAU_PROFILING
#include "TAU.h"
#endif

#include "version.h"

using namespace std;

int
main( int argc, char* argv[] ) 
{
    #ifdef USE_TAU_PROFILING
    TAU_PROFILE_INIT( argc, argv );
    TAU_PROFILE( "main()","int(int, char*)", TAU_DEFAULT);
    #endif
    // preset the values for non-MPI versions
    int 		rank = 0;
    int 		n_cpus = 1;

    #ifdef __DISTRIBUTED
    MPI_Init( &argc, &argv );
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &n_cpus);

    cerr << argv[ 0 ] << ": cpus=" << n_cpus << " rank=" << rank << endl;
    #endif 

    #ifdef USE_TAU_PROFILING
    TAU_PROFILE_SET_NODE( rank );
    #endif

    #ifdef __DISTRIBUTED
    if ( rank == 0 )
    {
    #endif
        std::cout << "HiPerSAT version 0.9 (" << g_hipersatVersion 
            << ")" << std::endl;
    #ifdef __DISTRIBUTED
    }
    #endif

    time_t 	seconds;
    string    formFile;
    bool		abortEarly = false;

    if ( argc <= 1 )
    {
        cerr << "Usage: SignalCleaner <formFile>" << endl;
        abortEarly = true;
    }
    else
    {
        int		formArg = 1;
        bool	xMode = false;

        if ( argv[ 1 ][ 0 ] == '-' )
        {
            if ( argv[ 1 ][ 1 ] == 'X' )
            {
                xMode = true;
                formArg++;
            }
            else
            {
                cerr << "Unknown argument " << argv[ 1 ] << endl;
                abortEarly = true;
            }
        }

        if ( !abortEarly )
        {
            if ( formArg >= argc )
            {
                cerr << "Usage: SignalCleaner <formFile>" << endl;
                abortEarly = true;
            }
            else
            {
                formFile = argv[ formArg ];
            }

            if ( xMode )
            {
                cerr << "xMode...Using form: " << formFile << endl;
                
                char        buff[ 1000 ];
                
                if ( getcwd( buff, sizeof( buff ) - 1 ) != 0 )
                {
                    cerr << "xMode...cwd: " << buff << endl;
                }
                
                abortEarly = true;
            }
        }

        if ( abortEarly )
        {
            #ifdef __DISTRIBUTED 
            MPI_Finalize(); 
            return EXIT_FAILURE; 
            #endif
        }
      
        #ifdef __PERFORMANCE_TIMING 
        struct tms start, interim, stop; 
        clock_t    begin, load_time, read_time, preprocess_time, process_time, end; 
        int        status;
        long       clktck;

        if (rank == 0)
        {
            begin = times(&start);
        }
        #endif 


        #ifdef __DISTRIBUTED
        const size_t    c_maxFormSize = 20000;
        char            formBuf[ c_maxFormSize ];

        if ( rank == 0 )
        {
            ifstream        formStream( formFile.c_str() );
            char*           insertHere = formBuf;

            while ( formStream )
            {
                string  line;
                getline( formStream, line );
                                                                                    //
                if ( insertHere - formBuf >= c_maxFormSize - line.length() ) 
                {  
                    cerr << "Input form: " << formFile << " is too large" << endl;  //
                    exit( 1 ); 
                }
                memcpy( insertHere, line.c_str(), line.length() );
                insertHere += line.length();
                *insertHere++ = '\n';
            }
                        
            *insertHere = 0;
                           
            formStream.close();
        }                     
                                                                                    //
        MPI_Bcast( formBuf, c_maxFormSize, MPI_CHAR, 0, MPI_COMM_WORLD );
        istringstream   formStrStream( formBuf );
                                                
        // read processing form 
                               
        ProcessReader*      process_reader = 
                            new ProcessReader( rank, n_cpus, formStrStream );
        #else 

        ifstream        x( formFile.c_str() );
        
        // read processing form ------------------------------------------------------
        ProcessReader*    process_reader = new ProcessReader( rank, n_cpus, x );

        #endif

        nic_matrix* data = process_reader->data;

        Preprocessor * preprocessor  = process_reader->preprocessor;
        Processor * processor = process_reader->process;
        Output * output = process_reader->output;

        if (process_reader->validate) 
        {
            process_reader->validate_load();
        }

        // cleanup form readers ------------------------------------------------------

        delete process_reader;

        //
        // Hack to seed random number generator with entropy
        //

        int     t = (int) time( 0 );
        
        #ifdef __DISTRIBUTED
        // give everyone the same time seed
        // this won't guarantee that the generators don't get out of step, though
        MPI_Bcast( &t, 1, MPI_INT, 0, MPI_COMM_WORLD );
        #endif

        cout << "srand( " << t << " )" << endl;
        srand( (int) t );

        #ifdef __PERFORMANCE_TIMING
        if (rank == 0)
        {            
            load_time = times(&start);
        } 
        #endif

        if(rank==0) 
        {
            seconds = time(NULL);
        }
        if (preprocessor != NULL) 
        {
            preprocessor->run(); 
        }
        if(rank==0) 
        {
            cout<<(time(NULL) - seconds)<<" PREPROCESSOR TIME\n";
            seconds = time(NULL);
        }

        #ifdef __PERFORMANCE_TIMING 
        if (rank == 0)
        {
            preprocess_time = times(&start); 
        }
        #endif 

        if (processor != NULL) 
        {
            processor->run();
        }

        if(rank==0) 
        {
            seconds = time(NULL) - seconds;
            cout << "SECONDS FOR EXECUTION : " << seconds << endl;
        }
        
        #ifdef __PERFORMANCE_TIMING
        if (rank == 0)            
        {                        
            process_time = times(&start);
        }                               
        #endif 
       
        if (output != NULL)
        {
            output->setup(preprocessor, processor);
            output->run(rank); //passing rank, now each node calls the output run procedure.
        }
        delete processor;
        delete output;
        delete preprocessor;

        #ifdef __PERFORMANCE_TIMING
        if (rank == 0)
        {
            end = times(&stop);
            clktck = sysconf(_SC_CLK_TCK);

            std::cout << (load_time - begin)/(double)clktck << "\t";
            std::cout << (preprocess_time - load_time)/(double)clktck << "\t";
            std::cout << (process_time - preprocess_time)/(double)clktck << "\t";
            std::cout << (end - process_time)/(double)clktck << "\t" << std::endl;
        }
        #endif

        #ifdef __DISTRIBUTED 
        MPI_Finalize();
        #endif 
    }

    return 0;
}

