#ifndef SPHERING_MATRIX_H_UNIVERSITY_OF_OREGON_NIC
#define SPHERING_MATRIX_H_UNIVERSITY_OF_OREGON_NIC

#include "sphering_matrix.h"
#include "nic_matrix.h"
#include "nic_vector.h"
#include "MatrixOperations.h"
#include "eigenstuff.h"
#include "LAPACKWrapper.h"
#include "covariance.h"
#include "center_data.h"
#include "average.h"

#include <algorithm>

#ifdef __OPENMP_
#include "omp.h"
#endif

#include <math.h>

#include <iostream> // for debugging

using namespace std;

// this method assumes that the data has been centered
void computeSpheringMatrix( 
    nic_matrix* centeredData, 
    nic_matrix* spheringMatrix,
    bool useInfomax )
{
    int rows = centeredData->rows;
    nic_matrix covrnce( rows, rows );
    nic_vector averageRows( rows );

    covariance( centeredData, &covrnce, &averageRows );

    // compute the eigenvalues and eigenvectors = V
    nic_matrix eigenvectors( rows, rows );
    nic_vector eigenvalues( rows );

    eigenvectors = covrnce; // make a copy of the data to be analyzed
    compute_eigenstuff( &eigenvectors, &eigenvalues );

    // compute lambda^(-1/2) = l
    for ( int i = 0; i < eigenvalues.n; ++i )
    {
        eigenvalues.data[i] = 1.0/sqrt(eigenvalues.data[i]);
        if ( useInfomax )
        {
            eigenvalues.data[i] *= 2;
        }
    }

    // compute l * V'
    bool transpose = true;
    multiplyAdiag( eigenvalues, eigenvectors, *spheringMatrix, transpose );

    // Infomax does additional computation; 2*V*l*V'
    if ( useInfomax )
    {
        nic_matrix infomaxSphering( rows, rows ); // = new nic_matrix( rows, rows );
        multiply( eigenvectors, *spheringMatrix, infomaxSphering );
        (*spheringMatrix) = infomaxSphering;
    }
}

void 
compute_sphering(nic_matrix* cov, nic_matrix * sph, 
			     nic_matrix* q, nic_vector* lambda,
			     bool useInfomaxSphering )
{
    int n = cov->rows;
    double* d = NULL;

    // compute eigenstuff

    // this operation performs a deep copy of the matrix
    *q = *cov;

    compute_eigenstuff( q, lambda );

    // compute lambda -1/2
    d = lambda->data;
    for (int i = 0; i < n; i++) 
    {
        d[i] = 1/sqrt(d[i]);
    }


    // compute sphering matrix S = L^{-1/2} Q^T
    double* 	l = lambda->data;
    double* 	s = sph->data;
    double* 	q_ = q->data;

    double * q_ptr = q_;


    for (int i = 0; i < n; i++) 
    {
        q_ptr = q_ + i;
        for (int j = 0; j < n; j++) 
        {
            *s++ = *q_ptr * l[j];
            q_ptr += n;
        }
    }

	if ( useInfomaxSphering )
	{
		//
		//	The Infomax sphering algorithm is different from that used by FastICA.
		//
		//	Infomax creates the sph matrix as:
		//		xxx = sqrtm( yyy )
		//
		//	Whereas FastICA computes it as:
		//		www = ????
		//
		
		//
		//	At this point, we have computed S = 
		
	    //	cerr << endl << "Using Infomax sphering..." << endl;
		
		double alpha = 2.0;
		double beta = 0.0;

  		nic_matrix* 	tmpSph = new nic_matrix( *sph );
		
		s = sph->data;

		dgemm_( "N", "N", &n, &n, &n, &alpha, 
            q_, &n, tmpSph->data, &n, &beta, s, &n );
		
		delete tmpSph;
    }
	
    //Infomax sph=2*q*(fastica sph).
};

#endif
// SPHERING_MATRIX_H_UNIVERSITY_OF_OREGON_NIC
