#ifndef WHITEN_DATA_CPP_UNIVERSITY_OF_OREGON_NIC
#define WHITEN_DATA_CPP_UNIVERSITY_OF_OREGON_NIC

#include "whiten_data.h"

using namespace std;

extern void covariance( nic_matrix*, nic_matrix*, nic_vector *);
extern void average( nic_matrix*, nic_vector* );
extern void compute_sphering( nic_matrix*, nic_matrix*, 
			     nic_matrix*, nic_vector*, bool useInfomaxSphering );
extern void center_data( nic_matrix*, nic_vector* );


//------------------------------------------------------------

void 
whiten_data(nic_matrix* data, nic_vector* avg,
    nic_matrix* cov, nic_matrix * q,
    nic_matrix* sph, nic_vector* lambda,
    bool compute_sph,
    bool	useInfomaxSphering ) 
{
    if (compute_sph) 
    {
        average( data, avg );
        center_data( data, avg );
        covariance(data, cov, avg);
        compute_sphering(cov, sph, q, lambda, useInfomaxSphering);
    } else 
    {
        average(data, avg);
        center_data(data, avg);
    }

    whiten(sph, data);
};

void
whiten(nic_matrix* sphere, nic_matrix * matrix) 
{
    int n_channels = matrix->rows;
    double * temp = (double *)malloc(sizeof(double) * n_channels);
    double * data = matrix->data;
    double * sphering_matrix = sphere->data;
    int observations = matrix->columns;

    double * sample_point = NULL;
    int i = 0, j = 0, k = 0;

    // start shared memory block
    // unroll this loop

    // Whiten the data
    for (i = 0; i < observations; i++) 
    {
        sample_point = data + i*n_channels;

        //    printf("r[%d] ", i);
        for (j = 0; j < n_channels; j++) 
        {
            //      printf("%g ", sample_point[k]);
            temp[j] = 0;
        }
        //    printf("\n");

        for (j = 0; j < n_channels; j++) 
        {
            for (k = 0; k < n_channels; k++) 
            {
                temp[j] += sphering_matrix[j + k * n_channels] * sample_point[k];
            }
        }

        for (k = 0; k < n_channels; k++) 
        {
            sample_point[k] = temp[k];
        }
    }
    free (temp);

    // end of shared memory block
};

// compute averge vector a' and covariance matrix of data' c'
// print data', avg, cov, sph, q and lamba to:
//     validate_whiten.dat -- unwhitened data   
//     validate_whiten.avg -- average vector    
//     validate_whiten.cov -- covariance matrix 
//     validate_whiten.sph -- sphering matrix   
//     validate_whiten.q   -- Q                 
//     validate_whiten.lam -- lambda vector     
//     validate_whiten.avg_p -- average vector    
//     validate_whiten.cov_p -- covariance matrix 
void
validate_whitening(nic_matrix* data, nic_vector* avg,
   nic_matrix* cov, nic_matrix* q,
   nic_matrix* sph, nic_vector* lambda,
   const std::string&             outputDir )
{
    int n = cov->rows;

    BinaryVectorWriter bavg_writer(
        avg, ( outputDir + "bvalidate_whitening.avg" ).c_str(),
        BinaryVectorWriter::c_fileType_binaryNativeEndian );

    BinaryMatrixWriter bcov_writer(
        cov, ( outputDir + "bvalidate_whitening.cov" ).c_str(),
        0, 0, 0, BinaryMatrixWriter::c_fileType_binaryNativeEndian );

    BinaryMatrixWriter bsph_writer(
        sph, ( outputDir + "bvalidate_whitening.sph" ).c_str(),
        0, 0, 0, BinaryMatrixWriter::c_fileType_binaryNativeEndian );

    BinaryMatrixWriter bq_writer(
        q, ( outputDir + "bvalidate_whitening.q" ).c_str(),
        0, 0, 0, BinaryMatrixWriter::c_fileType_binaryNativeEndian );

    BinaryVectorWriter blam_writer(
        lambda, ( outputDir + "bvalidate_whitening.lam" ).c_str(),
        BinaryVectorWriter::c_fileType_binaryNativeEndian );

    ASCIIVectorWriter aavg_writer(
        avg, ( outputDir + "avalidate_whitening.avg" ).c_str());

    ASCIIMatrixWriter acov_writer(
        cov, ( outputDir + "avalidate_whitening.cov" ).c_str());

    ASCIIMatrixWriter asph_writer(
        sph, ( outputDir + "avalidate_whitening.sph" ).c_str());

    ASCIIMatrixWriter aq_writer(
        q, ( outputDir + "avalidate_whitening.q" ).c_str());

    ASCIIVectorWriter alam_writer(
        lambda, ( outputDir + "avalidate_whitening.lam" ).c_str());

    nic_matrix* cov_prime = new nic_matrix(n, n);
    nic_vector* avg_prime = new nic_vector(n);
    covariance(data, cov_prime, avg_prime);

    BinaryVectorWriter bavgP_writer(
        avg_prime, ( outputDir + "bvalidate_whitening.avg_p" ).c_str(),
        BinaryVectorWriter::c_fileType_binaryNativeEndian );

    BinaryMatrixWriter bcovP_writer(
        cov_prime, ( outputDir + "bvalidate_whitening.cov_p" ).c_str(),
        0, 0, 0, BinaryMatrixWriter::c_fileType_binaryNativeEndian );

    ASCIIVectorWriter aavgP_writer(
        avg_prime, ( outputDir + "avalidate_whitening.avg_p" ).c_str());

    ASCIIMatrixWriter acovP_writer(
        cov_prime, ( outputDir + "avalidate_whitening.cov_p" ).c_str());

    delete avg_prime;
    delete cov_prime;
}


#ifdef DELETEME
void
store_whitened_data( nic_matrix* data, 
    const std::string& outputFilename )
{
    BinaryMatrixTWriter< double > bwhitened_data( data, outputFilename.c_str() );
}
#endif


#endif
// WHITEN_DATA_CPP_UNIVERSITY_OF_OREGON_NIC
