#ifndef TESTSPHERING_CPP_UNIVERSITY_OF_OREGON_NIC
#define TESTSPHERING_CPP_UNIVERSITY_OF_OREGON_NIC

#include "SimpleCLParser.cpp"
#include "nic_matrix.cpp"
#include "nic_vector.cpp"
#include "MatrixReader.cpp"
#include "DataReader.cpp"
#include "MatrixOperations.cpp"
#include "FastICASettings.h"
#include "FastICANew.cpp"
#include "ContrastFunctions.cpp"
#include "MPIWrapper.cpp"
#include "MersenneTwister.cpp"
#include "MPIWrapper.cpp"

#include <iostream>
#include <fstream>
#include <string>

using namespace std;

void setupParser( SimpleCLParser& parser )
{
    bool isRequired = true;
    bool isNotRequired = false;

    parser.addOption( SimpleCLParser::STRING, "-icen", "centered data", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-isph", "sphering matrix", isRequired );

    parser.addOption( SimpleCLParser::INTEGER, "-c", "number of channels", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-s", "number of samples", isRequired );
    parser.addOption( SimpleCLParser::DOUBLE, "-tol", "convergence tolerance", isRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-maxIter", "maximum number of iterations", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-contrast", "contrast function (cubic, hyptan, gaussian)", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-init", "initialization type (identity, random, user)", isNotRequired );
    parser.addOption( SimpleCLParser::STRING, "-initFile", "file for user initialization", isNotRequired );
    parser.addOption( SimpleCLParser::INTEGER, "-maxRetries", "maximum number of retries", isRequired );
    parser.addOption( SimpleCLParser::STRING, "-mixingMatrix", "precomputed mixing matrix to test against", isRequired );

    parser.addOption( SimpleCLParser::FLAG, "-h", "print usage", isNotRequired );
    parser.parse();

    if ( parser.readFlagOption( "-h" ) )
    {
        cout << parser.usage() << endl;
        exit( 0 );
    }

    parser.checkRequired();
}

void loadSettings( FastICASettings& settings, SimpleCLParser& parser )
{
    settings.convergenceTolerance = parser.readDoubleOption( "-tol" );
    settings.maximumIterations = parser.readIntOption( "-maxIter" );
    string cf = parser.readStringOption( "-contrast" );
    if ( cf == "cubic" )
    {
        settings.contrastFunction = ContrastFunctions::CUBIC;
    }
    else if ( cf == "hyptan" )
    {
        settings.contrastFunction = ContrastFunctions::HYPERBOLIC_TAN;
    }
    else if ( cf == "gaussian" )
    {
        settings.contrastFunction = ContrastFunctions::GAUSSIAN;
    }
    else
    {
        cerr << "unknown contrast function " << cf << endl;
        cerr << "exiting." << endl;
        exit(1);
    }
    string initType = parser.readStringOption( "-init" );
    if ( initType == "identity" )
    {
        settings.initializationType = FastICASettings::IDENTITY;
    }
    else if ( initType == "random" )
    {
        settings.initializationType = FastICASettings::RANDOM;
    }
    else if ( initType == "user" )
    {
        settings.initializationType = FastICASettings::USER_SPECIFIED;
    }
    else
    {
        cerr << "unknown initialization type " << initType << endl;
        cerr << "exiting" << endl;
        exit(1);
    }
    settings.userInitializationFile = parser.readStringOption( "-initFile" );
    settings.userInitializationFileFormat = MatrixReader::BIG;
    settings.maximumRetries = parser.readIntOption( "-maxRetries" );
}

int main( int argc, char** argv )
{
#ifdef __DISTRIBUTED
    MPI_Init( &argc, &argv );
#endif
    int rank = 0;
    int size = 1;

    // set up the parser
    SimpleCLParser clParser( argc, argv );
    setupParser( clParser );

    // read the data
    int channels = clParser.readIntOption( "-c" );
    int samples = clParser.readIntOption( "-s" );

    nic_matrix data;
    nic_matrix sphering;
    nic_matrix mixing;
    nic_matrix invMixing;
    
    bool loaded;

    loaded = MatrixReader::loadMatrix(
        data,
        MatrixReader::BIG,
        channels,
        samples,
        clParser.readStringOption( "-icen" ) );
    if ( !loaded )
    {
        cerr << "Failed to load " << clParser.readStringOption( "-icen" )
            << endl;
    }

    loaded = MatrixReader::loadMatrix(
        sphering,
        MatrixReader::BIG,
        channels,
        channels,
        clParser.readStringOption( "-isph" ) );
    if ( !loaded )
    {
        cerr << "Failed to load " << clParser.readStringOption( "-isph" )
            << endl;
    }

    nic_matrix realMixing;

    loaded = MatrixReader::loadMatrix(
        realMixing,
        MatrixReader::BIG,
        channels,
        channels,
        clParser.readStringOption( "-mixingMatrix" ) );
    if ( !loaded )
    {
        cerr << "Failed to load " << clParser.readStringOption( "-mixingMatrix" )
            << endl;
    }

    // load the settings into the settings object
    FastICASettings settings;
    loadSettings( settings, clParser );

    // run our method
    cout << "Testing FastICA" << endl;
    FastICANew fastIca( settings, &data, 123456UL, true );
    fastIca.setWhiteningMatrix( &sphering );
    fastIca.runFastICA();
    
    fastIca.mixingMatrix( mixing );
    invert( mixing, invMixing );

    double error;
    double tolerance = 1e-6;
    for ( int i = 0; i < mixing.rows; ++i )
    {
        for ( int j = 0; j < mixing.columns; ++j )
        {
            error = fabs( invMixing(i,j) - realMixing(j,i) );
            if ( !(error < tolerance ))
            {
                cout  << i << " " << invMixing(i,j) << " " << realMixing(j,i) <<
                    " " << error << endl;
            }
            assert( fabs( invMixing(i,j) - realMixing(j,i) ) < tolerance );
        }
    }

    cout << "   FastICA passed" << endl;

    return 0;
#ifdef __DISTRIBUTED
    MPI_Finalize();
#endif
}

#endif
// TESTSPHERING_CPP_UNIVERSITY_OF_OREGON_NIC
